(* Author: Vincent Cremet *)
(* Date: December 2006 *) 
(* File: Auxiliary.v *)
(* Description: Auxiliary definitions for FGJ-omega *)

(** * Auxiliary definitions for FGJ-omega *)


Require Import List.
Require Import Relations.
Require Import Arith.

Require Import MatchCompNat.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Substitutions_proofs.v *)
Require Substitutions_proofs.
Module My_Substitutions_proofs := Substitutions_proofs.SetProgram(My_Program).
Import My_Substitutions_proofs.

Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(******************************************************************)

Require Import Peano_dec.

Lemma Type_eqdec:
  forall (t u: TypeS),
    { t = u } + { t <> u }
.

Proof.
decide equality.
apply eq_nat_dec.
apply ClassSym_eqdec.
apply eq_nat_dec.
Qed.

(*****************************************************************)

Lemma TypeS_at_is_a_partial_function:
  forall (ts t t': TypeS) (i: nat),
    (TypeS_at ts i t) ->
    (TypeS_at ts i t') ->
    (t = t')
.

Proof.
intros ts t t' i H1 H2. generalize t' H2. clear t' H2. induction H1.
intros. inversion H2. trivial.
intros. inversion H2. apply IHTypeS_at. trivial.
Qed.

Lemma TypeS_at_of_typeSubst:
  forall (ts ti R: TypeS) (X: TypeVar) (i: nat),
  (TypeS_at ts i ti) ->
  (TypeS_at (typeSubst_in_type ts X R) i (typeSubst_in_type ti X R))
.

Proof.
intros ts ti R X i H. induction H.
simpl. apply TypeS_at_head.
simpl. apply TypeS_at_tail. apply IHTypeS_at.
Qed.

Lemma TypeS_at_of_lift:
  forall (ts ti: TypeS) (n k i: nat),
  (TypeS_at ts i ti) ->
  (TypeS_at (lift_at_depth_in_type ts n k) i (lift_at_depth_in_type ti n k))
.

Proof.
intros ts ti n k i H. induction H.
simpl. apply TypeS_at_head.
simpl. apply TypeS_at_tail. apply IHTypeS_at.
Qed.

Lemma Substitution_in_reduction:
  forall (R t t': TypeS),
    (RedType t t') ->
    forall (k: nat),
      (RedType (typeSubst_in_type t k R) (typeSubst_in_type t' k R))
.

Proof.
intros R t t' H. induction H.
(* RT-proj-sub *)
intro. simpl. apply RT_proj_sub. apply IHRedType.
(* RT-proj-redex *)
intro. simpl.
apply RT_proj_redex. apply TypeS_at_of_typeSubst. trivial.
(* RT-fun-left *)
intro. simpl. apply RT_fun_left. apply IHRedType.
(* RT-fun-right *)
intro. simpl. apply RT_fun_right. apply IHRedType.
(* RT-apply-redex *)
intro. simpl.
replace  (typeSubst_in_type (typeSubst_in_type T 0 R0) k R) with
  (typeSubst_in_type (typeSubst_in_type T (S k) R) 0 (typeSubst_in_type R0 k R)).
apply RT_apply_redex.
pattern k at 3. replace (k) with (0 + k).
rewrite Permuting_type_substitutions. trivial. trivial.
(* RT-apply-left *)
intro. simpl. apply RT_apply_left. apply IHRedType.
(* RT-apply-right *)
intro. simpl. apply RT_apply_right. apply IHRedType.
(* RT-tuple *)
intro. simpl. apply RT_tuple. apply IHRedType.
(* RT-param-left *)
intro. simpl. apply RT_param_left. apply IHRedType.
(* RT-param-right *)
intro. simpl. apply RT_param_right. apply IHRedType.
(* RT-section *)
intro. simpl. apply RT_section. apply IHRedType.
(* RT-some *)
intro. simpl. apply RT_some. apply IHRedType.
(* RT-cons-head *)
intro. simpl. apply RT_cons_head. apply IHRedType.
(* RT-cons-tail *)
intro. simpl. apply RT_cons_tail. apply IHRedType.
Qed.

Inductive RedTypeStar: TypeS -> TypeS -> Prop :=
  | RedType_step: forall (t t': TypeS), (RedType t t') -> (RedTypeStar t t')
  | RedType_refl: forall (t: TypeS), (RedTypeStar t t)
  | RedType_trans: forall (t u t': TypeS), (RedTypeStar t u) -> (RedTypeStar u t') ->
    (RedTypeStar t t')
.

Inductive RedTypeEnvStar: TypeEnv -> TypeEnv -> Prop :=
  | RedTypeEnv_step: forall (D D': TypeEnv), (RedTypeEnv D D') -> (RedTypeEnvStar D D')
  | RedTypeEnv_refl: forall (D: TypeEnv), (RedTypeEnvStar D D)
  | RedTypeEnv_trans: forall (D1 D2 D3: TypeEnv), (RedTypeEnvStar D1 D2) ->
    (RedTypeEnvStar D2 D3) -> (RedTypeEnvStar D1 D3)
.

(* Definition RedTypeStar: TypeS -> TypeS -> Prop := clos_refl_trans TypeS RedType. *)

Lemma RTS_proj_sub:
  forall (R R': TypeS) (i: nat),
    (RedTypeStar R R') ->
    (RedTypeStar (Type_proj R i) (Type_proj R' i))
.

Proof.
intros R R' i H. induction H.
(* step *)
apply RedType_step. apply RT_proj_sub. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_fun_left:
  forall (Pi Pi' T: TypeS),
    (RedTypeStar Pi Pi') ->
    (RedTypeStar (Type_fun Pi T) (Type_fun Pi' T))
.

Proof.
intros Pi Pi' T H. induction H.
(* step *)
apply RedType_step. apply RT_fun_left. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_fun_right:
  forall (Pi T T': TypeS),
    (RedTypeStar T T') ->
    (RedTypeStar (Type_fun Pi T) (Type_fun Pi T'))
.

Proof.
intros Pi T T' H. induction H.
(* step *)
apply RedType_step. apply RT_fun_right. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_apply_left:
  forall (K K' R: TypeS),
    (RedTypeStar K K') ->
    (RedTypeStar (Type_apply K R) (Type_apply K' R))
.

Proof.
intros K K' R H. induction H.
(* step *)
apply RedType_step. apply RT_apply_left. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_apply_right:
  forall (K R R': TypeS),
    (RedTypeStar R R') ->
    (RedTypeStar (Type_apply K R) (Type_apply K R'))
.

Proof.
Proof.
intros K R R' H. induction H.
(* step *)
apply RedType_step. apply RT_apply_right. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_tuple:
  forall (Ks Ks': TypeS),
    (RedTypeStar Ks Ks') ->
    (RedTypeStar (Type_tuple Ks) (Type_tuple Ks'))
.

Proof.
Proof.
intros Ks Ks' H. induction H.
(* step *)
apply RedType_step. apply RT_tuple. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_param_left:
  forall (Pi Pi' T: TypeS),
    (RedTypeStar Pi Pi') ->
    (RedTypeStar (Type_param Pi T) (Type_param Pi' T))
.

Proof.
Proof.
intros Pi Pi' T H. induction H.
(* step *)
apply RedType_step. apply RT_param_left. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_param_right:
  forall (Pi T T': TypeS),
    (RedTypeStar T T') ->
    (RedTypeStar (Type_param Pi T) (Type_param Pi T'))
.

Proof.
Proof.
intros Pi T T' H. induction H.
(* step *)
apply RedType_step. apply RT_param_right. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_section:
  forall (Ps Ps': TypeS),
    (RedTypeStar Ps Ps') ->
    (RedTypeStar (Type_section Ps) (Type_section Ps'))
.

Proof.
Proof.
intros Ps Ps' H. induction H.
(* step *)
apply RedType_step. apply RT_section. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_some:
  forall (T T': TypeS),
    (RedTypeStar T T') ->
    (RedTypeStar (Type_some T) (Type_some T'))
.

Proof.
Proof.
intros T T' H. induction H.
(* step *)
apply RedType_step. apply RT_some. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_cons_head:
  forall (t t' ts: TypeS),
    (RedTypeStar t t') ->
    (RedTypeStar (Type_cons t ts) (Type_cons t' ts))
.

Proof.
Proof.
intros t t' ts H. induction H.
(* step *)
apply RedType_step. apply RT_cons_head. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma RTS_cons_tail:
  forall (t ts ts': TypeS),
    (RedTypeStar ts ts') ->
    (RedTypeStar (Type_cons t ts) (Type_cons t ts'))
.

Proof.
Proof.
intros t ts ts' H. induction H.
(* step *)
apply RedType_step. apply RT_cons_tail. trivial.
(* refl *)
apply RedType_refl.
(* trans *)
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma Reduction_in_lifting:
  forall (t t': TypeS) (n: nat),
    (RedType t t') ->
    forall (k: nat),
      (RedType (lift_at_depth_in_type t n k) (lift_at_depth_in_type t' n k))
.

Proof.
intros t t' n H. induction H.
(* RT-proj-sub *)
intro. simpl. apply RT_proj_sub. apply IHRedType.
(* RT-proj-redex *)
intro. simpl. apply RT_proj_redex.
apply TypeS_at_of_lift. trivial.
(* RT-fun-left *)
intro. simpl. apply RT_fun_left. apply IHRedType.
(* RT-fun-right *)
intro. simpl. apply RT_fun_right. apply IHRedType.
(* RT-apply-redex *)
intro. simpl.
replace  (lift_at_depth_in_type (typeSubst_in_type T 0 R) n k) with
(typeSubst_in_type (lift_at_depth_in_type T n (S k)) 0 (lift_at_depth_in_type R n k)).
apply RT_apply_redex.
pattern k at 1. replace (k) with (k + 0).
pattern k at 3. replace (k) with (k + 0).
rewrite Permuting_type_lifting_and_substitution_2_aux. trivial.
auto with arith. auto with arith.
(* RT-apply-left *)
intro. simpl. apply RT_apply_left. apply IHRedType.
(* RT-apply-right *)
intro. simpl. apply RT_apply_right. apply IHRedType.
(* RT-tuple *)
intro. simpl. apply RT_tuple. apply IHRedType.
(* RT-param-left *)
intro. simpl. apply RT_param_left. apply IHRedType.
(* RT-param-right *)
intro. simpl. apply RT_param_right. apply IHRedType.
(* RT-section *)
intro. simpl. apply RT_section. apply IHRedType.
(* RT-some *)
intro. simpl. apply RT_some. apply IHRedType.
(* RT-cons-head *)
intro. simpl. apply RT_cons_head. apply IHRedType.
(* RT-cons-tail *)
intro. simpl. apply RT_cons_tail. apply IHRedType.
Qed.

Lemma Reduction_in_substitution:
  forall (R R': TypeS),
    (RedType R R') ->
    forall (t: TypeS) (k: nat),
      (RedTypeStar (typeSubst_in_type t k R) (typeSubst_in_type t k R'))
.

Proof.
intros R R' H t. induction t.
(* proj *)
intro. simpl. apply RTS_proj_sub. apply IHt.
(* class *)
intro. simpl. apply RedType_refl.
(* fun *)
intro. simpl.
eapply RedType_trans. apply RTS_fun_left. apply IHt1.
eapply RedType_trans. apply RTS_fun_right. apply IHt2.
apply RedType_refl.
(* apply *)
intro. simpl.
eapply RedType_trans. apply RTS_apply_left. apply IHt1.
eapply RedType_trans. apply RTS_apply_right. apply IHt2.
apply RedType_refl.
(* var *)
rename t into X.
intro.  simpl.
unfold typeSubst_in_typeVar.
induction (lt_eq_lt_dec X k). induction a.
(* X < k *)
rewrite match_comp_nat3_lt.
rewrite match_comp_nat3_lt. apply RedType_refl. trivial. trivial.
(* X = K *)
rewrite match_comp_nat3_eq.
rewrite match_comp_nat3_eq.
apply RedType_step.
unfold lift_in_type.
apply Reduction_in_lifting.
trivial. trivial. trivial.
(* X > k *)
rewrite match_comp_nat3_gt.
rewrite match_comp_nat3_gt. apply RedType_refl. trivial. trivial.
(* tuple *)
intro. simpl. apply RTS_tuple. apply IHt.
(* param *)
intro. simpl.
eapply RedType_trans. apply RTS_param_left. apply IHt1.
eapply RedType_trans. apply RTS_param_right. apply IHt2.
apply RedType_refl.
(* section *)
intro. simpl. apply RTS_section. apply IHt.
(* none *)
intro. simpl. apply RedType_refl.
(* some *)
intro. simpl. apply RTS_some. apply IHt.
(* nil *)
intro. simpl. apply RedType_refl.
(* cons *)
intro. simpl.
eapply RedType_trans. apply RTS_cons_head. apply IHt1.
eapply RedType_trans. apply RTS_cons_tail. apply IHt2.
apply RedType_refl.
Qed.

Lemma Substitution_in_reduction_star:
  forall (R t t': TypeS) (k: nat),
    (RedTypeStar t t') ->
    (RedTypeStar (typeSubst_in_type t k R) (typeSubst_in_type t' k R))
.

Proof.
intros. induction H.
apply RedType_step. apply Substitution_in_reduction. trivial.
apply RedType_refl.
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma Reduction_in_substitution_star:
  forall (R R': TypeS) (t: TypeS) (k: nat),
    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type t k R) (typeSubst_in_type t k R'))
.

Proof.
intros. induction H.
apply Reduction_in_substitution. trivial.
apply RedType_refl.
eapply RedType_trans. apply IHRedTypeStar1. apply IHRedTypeStar2.
Qed.

Lemma Reduction_and_substitution_star:
  forall (t t' R R': TypeS) (k: nat),
    (RedTypeStar t t') ->
    (RedTypeStar R R') ->
    (RedTypeStar (typeSubst_in_type t k R) (typeSubst_in_type t' k R'))
.

Proof.
intros.
eapply RedType_trans.
apply Reduction_in_substitution_star. apply H0.
apply Substitution_in_reduction_star. trivial.
Qed.

Lemma Subclass_implies_class_expansion:
  forall (C C': ClassSym) (R: TypeS),
  (Subclass C C') ->
  exists R': TypeS,
    (Class_subtyping (Class_type C R) (Class_type C' R'))
.

Proof.
intros. generalize R. clear R. induction H.
(* refl *)
intros. exists R. apply CS_Refl.
(* extends *)
intros. destruct (IHSubclass (typeSubst_in_type Ks 0 R)).
exists x. eapply CS_Extends. apply H. simpl. trivial.
Qed.

End SetProgram.