(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: Normalization_proofs.v *)
(* Description: Proof of normalization of type reduction *)

(** * Proof of normalization of type reduction *)

Require Import List.
(* Require Import Arith. *)
(* Require Import Relations. *)

Require Import MatchCompNat.
Require Import Util.
Require Import WellFoundation.
Require Import SimplyTypedLambdaCalculus.
Require Import SimplyTypedLambdaCalculus_normalization.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Subtyping_proofs.v *)
Require Subtyping_proofs.
Module My_Subtyping_proofs := Subtyping_proofs.SetProgram(My_Program).
Import My_Subtyping_proofs.

Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(*******************************************************************)

Lemma wellKinded_implies_kind_of_exists:
  forall (D: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    ((c = TypeSection) \/ (c = TypeParam) \/ (c = TypeParamList)) ->
    exists k: KindS,
      (kind_of t k)
.

Proof.
intros D t c H. induction H.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1. discriminate H1.
intros. destruct H0. discriminate H0. destruct H0. discriminate H0. discriminate H0.
intros. destruct H3. discriminate H3. destruct H3. discriminate H3. discriminate H3.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1. discriminate H1.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1. discriminate H1.
intros. destruct H0. discriminate H0. destruct H0. discriminate H0. discriminate H0.
intros. destruct H1. discriminate H1. destruct H1.
assert (exists k: KindS, (kind_of Pi k)). tauto. destruct H2. exists (Kind_constr x).
apply Kind_of_param. trivial. discriminate H1.
intros. destruct H0. assert (exists k: KindS, (kind_of Ps k)). tauto. destruct H1.
exists (Kind_tuple x). apply Kind_of_section. trivial. destruct H0. discriminate H0.
discriminate H0.
intros. destruct H. discriminate H. destruct H. discriminate H. discriminate H.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1. discriminate H1.
intros. destruct H. discriminate H. destruct H. discriminate H. 
exists Kind_nil. apply Kind_of_paramList_nil.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1.
assert (exists k: KindS, (kind_of P k)). tauto.
assert (exists ks: KindS, (kind_of Ps ks)). tauto. destruct H2. destruct H3.
exists (Kind_cons x x0). apply Kind_of_paramList_cons; trivial.
intros. destruct H. discriminate H. destruct H. discriminate H. discriminate H.
intros. destruct H1. discriminate H1. destruct H1. discriminate H1. discriminate H1.
Qed.

(*******************************************************************)

Lemma Class_kind_existence:
  forall (C: ClassSym),
    exists k: KindS,
      (kind_of (classTypeParams C) k)
.

Proof.
intros.
destruct (classes_are_well_kinded C).
eapply wellKinded_implies_kind_of_exists. apply H. left. trivial.
Qed.

(******************************************************************)

Module My_Constants. (* : LambdaConstants. *)

Inductive LConst_definition: Set :=
  | LConst_class : ClassSym -> LConst_definition
  | LConst_none  : LConst_definition
  | LConst_nil   : LConst_definition
.

Definition LConst: Set := LConst_definition.

(** Types - S, T, U *)
Inductive LType: Set :=
  | LType_base : LType                    (*   *    *)
  | LType_fun  : LType -> LType -> LType  (* T -> U *)
  | LType_prod : LType -> LType -> LType  (* (T, U) *)
.

Fixpoint type_of_kind (k': KindS): LType :=
  match k' with
    | (Kind_tuple ks)  => (type_of_kind ks)
    | (Kind_constr kr) => (LType_prod (LType_fun (type_of_kind kr) LType_base)
                                      (LType_fun (type_of_kind kr) (type_of_kind kr)))
    | Kind_nil         => LType_base
    | (Kind_cons k ks) => (LType_prod (type_of_kind k) (type_of_kind ks))
  end
.

Inductive type_of (t: TypeS): LType -> Prop :=

  | Type_of_intro:
    forall (k: KindS),
      (kind_of t k) ->
      (type_of t (type_of_kind k))
.

Lemma Type_of_is_a_partial_function:
  forall (t: TypeS) (T1 T2: LType),
    (type_of t T1) ->
    (type_of t T2) ->
    T1 = T2
.

Proof.
intros. inversion H. inversion H0. replace k with k0. trivial.
eapply kind_of_is_a_partial_function; eauto.
Qed.

Inductive type_of_const_def: LConst -> LType -> Prop :=

  | type_of_const_class:
    forall (C: ClassSym) (k: KindS),
      (kind_of (classTypeParams C) k) ->
      (type_of_const_def (LConst_class C) (type_of_kind (Kind_constr k)))

  | type_of_const_none:
    (type_of_const_def LConst_none LType_base)

  | type_of_const_nil:
    (type_of_const_def LConst_nil LType_base)
.

Definition type_of_const := type_of_const_def.

End My_Constants.

(*******************************************************************)

Module My_LambdaCalculusNormalization := LambdaCalculusNormalization(My_Constants).

Import My_LambdaCalculusNormalization.
Import My_LambdaCalculus.
Import My_Constants.

(*******************************************************************)

Fixpoint LTerm_proj (t: LTerm) (i: nat) {struct i}: LTerm :=
  match i with
    | 0     => (LTerm_fst t)
    | (S n) => (LTerm_proj (LTerm_snd t) n)
  end
.

(** Translation from FGJ-omega types to lambda terms **)
Inductive trans: TypeS -> LTerm -> Prop :=

  | trans_class:
    forall (C: ClassSym),
      (trans (Type_class C) (LTerm_const (LConst_class C)))

  | trans_proj:
    forall (R: TypeS) (R': LTerm) (i: nat),
      (trans R R') ->
      (trans (Type_proj R i) (LTerm_proj R' i))

  | trans_fun:
    forall (Pi T: TypeS) (U: LType) (Pi' T': LTerm),
      (trans Pi Pi') ->
      (trans T T') ->
      (type_of Pi U) ->
      (trans (Type_fun Pi T) (LTerm_pair (LTerm_fun U T') Pi'))

  | trans_var:
    forall (X: TypeVar),
      (trans (Type_var X) (LTerm_var X))

  | trans_tuple:
    forall (Ks: TypeS) (Ks': LTerm),
      (trans Ks Ks') ->
      (trans (Type_tuple Ks) Ks')

  | trans_apply:
    forall (K Ks: TypeS) (K' Ks': LTerm),
      (trans K K') ->
      (trans Ks Ks') ->
      (trans (Type_apply K Ks) (LTerm_app (LTerm_fst K')  Ks'))

  | trans_param:
    forall (Pi T_opt: TypeS) (U: LType) (Pi' T_opt': LTerm),
      (trans Pi Pi') ->
      (trans T_opt T_opt') ->
      (type_of Pi U) ->
      (trans (Type_param Pi T_opt) (LTerm_pair (LTerm_fun U T_opt') Pi'))

  | trans_section:
    forall (Ps: TypeS) (U: LType) (Ps': LTerm),
      (trans Ps Ps') ->
      (type_of (Type_section Ps) U) ->
      (trans (Type_section Ps) (LTerm_fun U Ps'))

  | trans_none:
    (trans Type_none (LTerm_const LConst_none))

  | trans_some:
    forall (T: TypeS) (T': LTerm),
      (trans T T') ->
      (trans (Type_some T) T')

  | trans_nil:
    (trans Type_nil (LTerm_const LConst_nil))

  | trans_cons:
    forall (t ts: TypeS) (t' ts':LTerm),
      (trans t t') ->
      (trans ts ts') ->
      (trans (Type_cons t ts) (LTerm_pair t' ts'))
.

Inductive trans_category (t: TypeS): TypeC -> LType -> Prop :=

  | transcat_TypeConstr:
    forall (k: KindS),
      (trans_category t (TypeConstr k) (type_of_kind k))

  | transcat_Type_: 
    (trans_category t Type_  LType_base)

  | transcat_TypeTuple: 
    forall (kr: KindS),
      (trans_category t (TypeTuple kr) (type_of_kind kr))

  | transcat_TypeParam:
    forall (k: KindS),
      (kind_of t k) ->
      (trans_category t TypeParam (type_of_kind k))

  | transcat_TypeSection:
    forall (k: KindS),
      (kind_of t k) ->
      (trans_category t TypeSection (LType_fun (type_of_kind k) (type_of_kind k)))

  | transcat_TypeOption: 
    (trans_category t TypeOption LType_base)
    
  | transcat_TypeParamList:
    forall (k: KindS),
      (kind_of t k) -> 
      (trans_category t TypeParamList (type_of_kind k))

  | transcat_TypeConstrList:
    forall (ks: KindS),
      (trans_category t (TypeConstrList ks) (type_of_kind ks))
.

(*******************************************************************)

Lemma R_proj_cong:
  forall (t t': LTerm) (i: nat),
    (RedTerm t t') ->
    (RedTerm (LTerm_proj t i) (LTerm_proj t' i))
.

Proof.
intros. generalize t t' H. clear t t' H. induction i.
(* 0 *)
intros. simpl. apply R_fst_cong; trivial.
(* S *)
intros. simpl. apply IHi. apply R_snd_cong; trivial.
Qed.

Lemma Rplus_proj_cong:
  forall (t t': LTerm) (i: nat),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_proj t i) (LTerm_proj t' i))
.

Proof.
intros. induction H. 
apply tc_step. apply R_proj_cong. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_fun:
  forall (t t': LTerm) (T: LType),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_fun T t) (LTerm_fun T t'))
.

Proof.
intros. induction H. 
apply tc_step. apply R_fun. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_app_left:
  forall (t t' u: LTerm),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_app t u) (LTerm_app t' u))
.

Proof.
intros. induction H. 
apply tc_step. apply R_app_left. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_app_right:
  forall (t u u': LTerm),
    (Rplus RedTerm u u') ->
    (Rplus RedTerm (LTerm_app t u) (LTerm_app t u'))
.

Proof.
intros. induction H. 
apply tc_step. apply R_app_right. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_pair_left:
  forall (t t' u: LTerm),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_pair t u) (LTerm_pair t' u))
.

Proof.
intros. induction H. 
apply tc_step. apply R_pair_left. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_pair_right:
  forall (t u u': LTerm),
    (Rplus RedTerm u u') ->
    (Rplus RedTerm (LTerm_pair t u) (LTerm_pair t u'))
.

Proof.
intros. induction H. 
apply tc_step. apply R_pair_right. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_fst_cong:
  forall (t t': LTerm),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_fst t) (LTerm_fst t'))
.

Proof.
intros. induction H. 
apply tc_step. apply R_fst_cong. trivial.
eapply tc_trans; eauto.
Qed.

Lemma Rplus_snd_cong:
  forall (t t': LTerm) (T: LType),
    (Rplus RedTerm t t') ->
    (Rplus RedTerm (LTerm_snd t) (LTerm_snd t'))
.

Proof.
intros. induction H. 
apply tc_step. apply R_snd_cong. trivial.
eapply tc_trans; eauto.
Qed.

(*******************************************************************)

Lemma typeSubst_of_LTerm_proj:
  forall (t s: LTerm) (i k: nat),
  (subst_in_term (LTerm_proj t i) k s) =
  (LTerm_proj (subst_in_term t k s) i)
.

Proof.
intros. generalize t. clear t. induction i.
intro. simpl. trivial.
intro. simpl. rewrite IHi. simpl. trivial.
Qed.

Lemma type_of_of_typeSubst:
  forall (t R: TypeS) (U: LType) (k: nat),
    (type_of t U) ->
    (type_of (typeSubst_in_type t k R) U)
.

Proof.
intros. induction H. apply Type_of_intro.
apply Substitution_preserves_kind_of. trivial.
Qed.

Lemma lift_of_LTerm_proj:
  forall (t: LTerm) (i k n: nat),
    (lift_at_depth_in_term (LTerm_proj t i) k n) =
    (LTerm_proj (lift_at_depth_in_term t k n) i)
.

Proof.
intros. generalize t. clear t. induction i.
intros. simpl. trivial.
intros. simpl. rewrite IHi. simpl. trivial.
Qed.

Lemma type_of_of_lift:
  forall (t: TypeS) (U: LType) (k n: nat),
    (type_of t U) ->
    (type_of (lift_at_depth_in_type t k n) U)
.

Proof.
intros. induction H. apply Type_of_intro.
apply Lifting_preserves_kind_of. trivial.
Qed.

Lemma trans_of_lift:
  forall (t: TypeS) (k n: nat) (t': LTerm),
    (trans t t') ->
    (trans (lift_at_depth_in_type t n k) (lift_at_depth_in_term t' k n))
.

Proof.
intros. generalize k. clear k. induction H.
(* class *)
intro. simpl. apply trans_class.
(* proj *)
intro. simpl. rewrite lift_of_LTerm_proj. apply trans_proj. auto.
(* fun *)
intro. simpl. apply trans_fun. auto. auto. apply type_of_of_lift; trivial.
(* var *)
intro. simpl. apply trans_var.
(* tuple *)
intro. simpl. apply trans_tuple. auto.
(* apply *)
intro. simpl. apply trans_apply; auto.
(* param *)
intro. simpl. apply trans_param. auto. auto. apply type_of_of_lift; trivial.
(* section *)
intro. simpl. apply trans_section. auto.
replace (Type_section (lift_at_depth_in_type Ps n (S k))) with
(lift_at_depth_in_type (Type_section Ps) n k). apply type_of_of_lift; trivial. trivial.
(* none *)
intro. simpl. apply trans_none.
(* some *)
intro. simpl. apply trans_some; auto. 
(* nil *)
intro. simpl. apply trans_nil.
(* cons *)
intro. simpl. apply trans_cons; auto.
Qed.

Lemma trans_of_typeSubst:
  forall (t R: TypeS) (k: nat) (t' R': LTerm),
    (trans t t') ->
    (trans R R') ->
    (trans (typeSubst_in_type t k R) (subst_in_term t' k R'))
.

Proof.
intros. generalize k. clear k. induction H.
(* class *)
intro. simpl. apply trans_class.
(* proj *)
intro. simpl. rewrite typeSubst_of_LTerm_proj. apply trans_proj. apply IHtrans.
(* fun *)
intro. simpl.
apply trans_fun. auto. auto.
apply type_of_of_typeSubst. trivial.
(* var *)
intro. simpl.
Require Import Arith.
destruct (lt_eq_lt_dec X k). destruct s.
 (* X < k *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_lt.
unfold subst_in_var. rewrite match_comp_nat3_lt.
apply trans_var. trivial. trivial.
 (* X = k *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_eq.
unfold subst_in_var. rewrite match_comp_nat3_eq.
unfold lift_in_type. unfold lift_in_term. apply trans_of_lift; auto.
trivial. trivial.
 (* X > k *)
unfold typeSubst_in_typeVar. rewrite match_comp_nat3_gt.
unfold subst_in_var. rewrite match_comp_nat3_gt.
apply trans_var. trivial. trivial.
(* tuple *)
intro. simpl. apply trans_tuple. auto.
(* apply *)
intro. simpl. apply trans_apply. auto. auto.
(* param *)
intro. simpl.
apply trans_param. auto. auto. apply type_of_of_typeSubst; trivial.
(* section *)
intro. simpl. 
apply trans_section. auto.
replace  (Type_section (typeSubst_in_type Ps (S k) R)) with
(typeSubst_in_type (Type_section Ps) k R).
apply type_of_of_typeSubst; trivial. trivial.
(* none *)
intro. simpl. apply trans_none.
(* some *)
intro. simpl. apply trans_some. auto.
(* nil *)
intro. simpl. apply trans_nil.
(* cons *)
intro. simpl. apply trans_cons. auto. auto.
Qed.

Lemma trans_is_a_partial_function:
  forall (t: TypeS) (u1 u2: LTerm),
    (trans t u1) ->
    (trans t u2) ->
    (u1 = u2)
.

Proof.
intros. generalize u2 H0. clear u2 H0. induction H.
intros. inversion H0. trivial.
intros. inversion H0. rewrite IHtrans with (u2 := R'0). trivial. trivial.
intros. inversion H2. rewrite IHtrans1 with (u2 :=Pi'0).
rewrite IHtrans2 with (u2 :=T'0).
replace U0 with U. trivial. eapply Type_of_is_a_partial_function; eauto.
trivial. trivial.
intros. inversion H0. trivial.
intros. inversion H0. apply IHtrans. trivial.
intros. inversion H1.
rewrite IHtrans1 with (u2 :=K'0).
rewrite IHtrans2 with (u2 :=Ks'0). trivial. trivial. trivial.
intros. inversion H2. 
rewrite IHtrans1 with (u2 :=Pi'0).
rewrite IHtrans2 with (u2 :=T_opt'0).
replace U0 with U. trivial. eapply Type_of_is_a_partial_function; eauto.
trivial. trivial.
intros. inversion H1.
rewrite IHtrans with (u2 :=Ps'0).
replace U0 with U. trivial. eapply Type_of_is_a_partial_function; eauto.
trivial.
intros. inversion H0. trivial.
intros. inversion H0. apply IHtrans. trivial.
intros. inversion H0. trivial.
intros. inversion H1.
rewrite IHtrans1 with (u2 :=t'0). 
rewrite IHtrans2 with (u2 :=ts'0). trivial. trivial. trivial.
Qed.

Lemma R_proj_redex:
  forall (i: nat) (Ks Ki: TypeS) (Ks' Ki': LTerm),
    (TypeS_at Ks i Ki) ->
    (trans Ks Ks') ->
    (trans Ki Ki') ->
    (Rplus RedTerm (LTerm_proj Ks' i) Ki')
.

Proof.
intros i Ks Ki Ks' Ki' H. generalize Ks' Ki'. clear Ks' Ki'.  induction H.
intros. simpl. apply tc_step. inversion H. 
replace t' with Ki'. apply R_fst_redex. eapply trans_is_a_partial_function; eauto.
intros. simpl. inversion H0. eapply tc_trans. apply tc_step.
apply R_proj_cong. apply R_snd_redex. apply IHTypeS_at. trivial. trivial.
Qed.

(*******************************************************************)

Lemma Reduction_preserves_type_of:
  forall (t t': TypeS) (T: LType),
    (type_of t T) ->
    (RedType t t') ->
    (type_of t' T)
.

Proof.
intros. induction H. apply Type_of_intro.
eapply Reduction_preserves_kind_of; eauto.
Qed.

(*******************************************************************)

Theorem Preservation_of_semantics:
  forall (t u: TypeS) (t' u': LTerm),
    (RedType t u) ->
    (trans t t') ->
    (trans u u') ->
    (Rplus RedTerm t' u')
.

Proof.
intros. generalize t' u' H0 H1. clear t' u' H0 H1. induction H.
(* R-proj-sub *)
intros. inversion H0. inversion H1.
apply Rplus_proj_cong. apply IHRedType; trivial.
(* R-proj-redex *)
intros. inversion H0. inversion H5. eapply R_proj_redex. apply H. trivial. trivial.
(* R-fun-left *)
intros. inversion H0. inversion H1.
replace U with U0.
replace T'0 with T'.
apply Rplus_pair_right. apply IHRedType; trivial.
eapply trans_is_a_partial_function; eauto.
assert (type_of Pi' U). eapply Reduction_preserves_type_of; eauto.
eapply Type_of_is_a_partial_function; eauto.
(* R-fun-right *)
intros. inversion H0. inversion H1.
replace Pi'0 with Pi'.
apply Rplus_pair_left.
replace U with U0.
apply Rplus_fun. eauto.
eapply Type_of_is_a_partial_function; eauto.
eapply trans_is_a_partial_function; eauto.
(* R-app-redex *)
intros. inversion H0. inversion H3.
replace u' with (subst_in_term T' 0 Ks').
eapply tc_trans. apply tc_step. apply R_app_left. apply R_fst_redex.
apply tc_step. apply R_app_redex.
eapply trans_is_a_partial_function.
apply trans_of_typeSubst. eauto. eauto. trivial.
(* R-app-left *)
intros. inversion H0. inversion H1.
replace Ks'0 with Ks'. apply Rplus_app_left. apply Rplus_fst_cong. eauto.
eapply trans_is_a_partial_function; eauto.
(* R-app-right *)
intros. inversion H0. inversion H1.
replace K'0 with K'. apply Rplus_app_right. eauto.
eapply trans_is_a_partial_function; eauto.
(* R-tuple *)
intros. inversion H0. inversion H1. eauto.
(* R-param-left *)
intros. inversion H0. inversion H1.
replace U with U0.
replace T_opt'0 with T_opt'.
apply Rplus_pair_right. eauto.
eapply trans_is_a_partial_function; eauto.
assert (type_of Pi' U). eapply Reduction_preserves_type_of; eauto.
eapply Type_of_is_a_partial_function; eauto.
(* R-param-right *)
intros. inversion H0. inversion H1.
replace Pi'0 with Pi'.
apply Rplus_pair_left.
replace U with U0.
apply Rplus_fun. eauto.
eapply Type_of_is_a_partial_function; eauto.
eapply trans_is_a_partial_function; eauto.
(* R-section *)
intros. inversion H0. inversion H1.
replace U with U0.
apply Rplus_fun. eauto.
assert (type_of (Type_section Ps') U). eapply Reduction_preserves_type_of.
apply H4. apply RT_section. trivial.
eapply Type_of_is_a_partial_function; eauto.
(* R-some *)
intros. inversion H0. inversion H1. eauto.
(* R-cons-head *)
intros. inversion H0. inversion H1.
replace ts'0 with ts'.
apply Rplus_pair_left. eauto.
eapply trans_is_a_partial_function; eauto.
(* R-cons-tail *)
intros. inversion H0. inversion H1.
replace t'1 with t'0.
apply Rplus_pair_right. eauto.
eapply trans_is_a_partial_function; eauto.
Qed.

(*******************************************************************)

Lemma WT_proj:
  forall (G: LEnv) (ks ki: KindS) (t: LTerm) (i: nat),
    (WellTyped G t (type_of_kind ks)) ->
    (KindS_at ks i ki) ->
    (WellTyped G (LTerm_proj t i) (type_of_kind ki))
.

Proof.
intros. generalize t H. clear t H. induction H0.
simpl. intros. eapply WT_fst; eauto.
simpl. intros. apply IHKindS_at. eapply WT_snd; eauto.
Qed.

(******************************************************************)

Inductive trans_typeEnv: TypeEnv -> LEnv -> Prop :=

  | trans_typeEnv_nil:
    (trans_typeEnv nil nil)

  | trans_typeEnv_snoc:
    forall (Pi: TypeS) (D: TypeEnv) (k: KindS) (D': LEnv),
    (kind_of Pi k) ->
    (trans_typeEnv D D') ->
    (trans_typeEnv (snoc D Pi) (snoc D' (type_of_kind k)))
 .

Lemma trans_typeEnv_and_TypeSection_at:
  forall (D: TypeEnv) (G: LEnv) (i: nat) (Pi: TypeS) (k: KindS),
    (TypeSection_at D i Pi) ->
    (trans_typeEnv D G) ->
    (kind_of Pi k) ->
    (List_from G i (type_of_kind k))
.

Proof.
intros. generalize G k H0 H1. clear G k H0 H1. induction H.
intros. inversion H0. absurd ((snoc D Pi) = nil). apply snoc_ind_aux_3. auto.
replace k with k0. apply List_from_tail. destruct (snoc_inject _ _ _ _ _ H).
eapply kind_of_is_a_partial_function. apply H2. rewrite H6. trivial.
intros. inversion H0. absurd ((snoc D Pi) = nil). apply snoc_ind_aux_3. auto.
apply List_from_head. apply IHTypeSection_at. replace D with D0. trivial.
destruct (snoc_inject _ _ _ _ _ H2). trivial. trivial.
Qed.

(******************************************************************)

Theorem Preservation_of_typing:
  forall (D: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    forall (G: LEnv),
      (trans_typeEnv D G) ->
      exists t': LTerm, exists c': LType,
        (trans t t') /\
        (trans_category t c c') /\
        (WellTyped G t' c')
.

Proof.
intros D t c H G Hy. generalize G Hy. clear G Hy. induction H.
(* WK-proj *)
intros.
destruct (IHType_WK _ Hy). destruct H1. destruct H1. destruct H2. rename x into R'.
rename x0 into c'.
exists (LTerm_proj R' i). exists (type_of_kind k'). split. apply trans_proj; trivial.
split. apply transcat_TypeConstr.
inversion H2. simpl in H4. rewrite <- H4 in H3.
eapply WT_proj; eauto.
(* WK_class *)
intros.
exists (LTerm_const (LConst_class C)).
exists (type_of_kind (Kind_constr k)). split.
apply trans_class. split.
apply transcat_TypeConstr.
apply WT_const. apply type_of_const_class. trivial.
(* WK_fun *)
intros.
destruct (IHType_WK1 _ Hy). destruct H3. destruct H3.  destruct H4. rename x into Pi'.
assert (trans_typeEnv (snoc D Pi) (snoc G (type_of_kind k))).
apply trans_typeEnv_snoc; trivial. rename H6 into Hy0.
destruct (IHType_WK2 _ Hy0). destruct H6. destruct H6. destruct H7. rename x into N'.
inversion H7. rewrite <- H10 in H8.
exists (LTerm_pair (LTerm_fun (type_of_kind k) N') Pi').
exists (type_of_kind (Kind_constr k)). split.
apply trans_fun. trivial. trivial. apply Type_of_intro; trivial.
split. apply transcat_TypeConstr.
simpl. inversion H4.
apply WT_pair.
apply WT_fun.
trivial. rewrite <- H12 in H5. replace k with k0. trivial.
eapply kind_of_is_a_partial_function; eauto.
(* WK_apply *)
intros.
destruct (IHType_WK1 _ Hy). destruct H1. destruct H1. destruct H2. rename x into K'.
destruct (IHType_WK2 _ Hy). destruct H4. destruct H4. destruct H5. rename x into R'.
exists (LTerm_app (LTerm_fst K') R').
exists (LType_base). split.
apply trans_apply; trivial. split. apply transcat_Type_.
inversion H2. rewrite <- H7 in H3. inversion H5. rewrite <- H9 in H6. simpl in H3.
eapply WT_app. eapply WT_fst. apply H3. trivial.
(* WK_var *)
intros.
exists (LTerm_var X). exists (type_of_kind k). split. apply trans_var. split.
apply transcat_TypeTuple. inversion H. clear D0 X0 H2 H3.
apply WT_var.
eapply trans_typeEnv_and_TypeSection_at. apply H1. trivial.
rewrite <- H4 in H0. eapply Lifting_preserves_kind_of_reverse. unfold lift_in_type in H0.
apply H0.
(* WK_tuple *)
intros.
destruct (IHType_WK _ Hy). destruct H0. destruct H0. destruct H1. rename x into Ks'.
exists Ks'. exists (type_of_kind (Kind_tuple k)). split.
apply trans_tuple; trivial. split. apply transcat_TypeTuple.
simpl. inversion H1.  rewrite <- H3 in H2. trivial.
(* WK_param *)
intros.
destruct (IHType_WK1 _ Hy). destruct H1. destruct H1.  destruct H2. rename x into Pi'.
inversion H2. rename H4 into Hyp1. rename H6 into Hyp2.
assert (trans_typeEnv (snoc D Pi) (snoc G (type_of_kind k))).
apply trans_typeEnv_snoc; trivial. rename H4 into Hy0.
destruct (IHType_WK2 _ Hy0). destruct H4. destruct H4.  destruct H5. rename x into N_opt'.
inversion H5. rewrite <- H8 in H6. rewrite <- Hyp2 in H3.
exists (LTerm_pair (LTerm_fun (type_of_kind k) N_opt') Pi').
exists (type_of_kind (Kind_constr k)). split.
apply trans_param; trivial. apply Type_of_intro; trivial. split.
apply transcat_TypeParam. apply Kind_of_param. trivial.
simpl. apply WT_pair. apply WT_fun.
trivial.
trivial.
(* WK_section *)
intros.
destruct (wellKinded_implies_kind_of_exists _ _ _ H). tauto. rename x into k.
rename H0 into Hyp1.
assert (trans_typeEnv (snoc D (Type_section Ps)) (snoc G (type_of_kind (Kind_tuple k)))).
apply trans_typeEnv_snoc; trivial. apply Kind_of_section; trivial. rename H0 into Hy0.
destruct (IHType_WK _ Hy0). destruct H0. destruct H0. destruct H1. rename x into Ps'.
inversion H1.
exists (LTerm_fun (type_of_kind (Kind_tuple k)) Ps').
exists (LType_fun (type_of_kind (Kind_tuple k)) (type_of_kind (Kind_tuple k))). split.
apply trans_section. trivial. apply Type_of_intro; trivial.
apply Kind_of_section. trivial. split.
apply transcat_TypeSection. apply Kind_of_section. trivial.
simpl. rewrite <- H5 in H2. apply WT_fun.
pattern k at 2. replace k with k0. trivial.
eapply kind_of_is_a_partial_function; eauto.
(* WK_none *)
intros.
exists (LTerm_const LConst_none). exists LType_base. split. apply trans_none.
split. apply transcat_TypeOption. apply WT_const. apply type_of_const_none. 
(* WK_some *)
intros.
destruct (IHType_WK _ Hy). destruct H1. destruct H1.  destruct H2. rename x into N'.
exists N'. exists LType_base. split.
apply trans_some; trivial. split. apply transcat_TypeOption.
inversion H2. rewrite <- H5 in H3. trivial.
(* WK_nil *)
intros.
exists (LTerm_const LConst_nil). exists LType_base. split.
apply trans_nil. split.
replace LType_base with (type_of_kind Kind_nil).
apply transcat_TypeParamList. apply Kind_of_paramList_nil. trivial.
apply WT_const. apply type_of_const_nil.
(* WT_cons *)
intros.
destruct (IHType_WK1 _ Hy). destruct H1. destruct H1. destruct H2. rename x into P'.
destruct (IHType_WK2 _ Hy). destruct H4. destruct H4. destruct H5. rename x into Ps'.
exists (LTerm_pair P' Ps').
inversion H2.  inversion H5. exists (type_of_kind (Kind_cons k k0)). split.
apply trans_cons. trivial. trivial. split. apply transcat_TypeParamList.
apply Kind_of_paramList_cons; trivial.
simpl. apply WT_pair. rewrite <- H9 in H3. trivial. rewrite <- H11 in H6. trivial.
(* WK_constr_nil *)
intros.
exists (LTerm_const LConst_nil).  exists LType_base. split.
apply trans_nil. split.
replace LType_base with (type_of_kind Kind_nil). apply transcat_TypeConstrList. trivial.
apply WT_const. apply type_of_const_nil.
(* WT_constr_cons *)
intros.
destruct (IHType_WK1 _ Hy). destruct H1. destruct H1.  destruct H2. rename x into K'.
destruct (IHType_WK2 _ Hy). destruct H4. destruct H4. destruct H5. rename x into Ks'.
exists (LTerm_pair K' Ks'). inversion H2. inversion H5.
exists (type_of_kind (Kind_cons k ks)). split.
apply trans_cons; trivial. split. apply transcat_TypeConstrList.
simpl. apply WT_pair. rewrite <- H7 in H3. trivial. rewrite <- H9 in H6. trivial.
Qed.

(*******************************************************************)

Lemma Preservation_of_semantics_cor1:
  forall (D: TypeEnv) (t: TypeS) (c: TypeC) (t': LTerm),
    (Type_WK D t c) ->
    (exists G: LEnv, (trans_typeEnv D G)) ->
    (trans t t') ->
    (Acc (symRel RedTerm) t') ->
    (Acc (symRel RedType) t)
.

Proof.
cut (
  forall (D: TypeEnv) (c: TypeC) (t': LTerm),
    (exists G: LEnv, (trans_typeEnv D G)) ->
    (Acc (symRel (Rplus RedTerm)) t') ->
    forall (t: TypeS),
      (Type_WK D t c) ->
      (trans t t') ->
      (Acc (symRel RedType) t)
).
intros. eapply H. eauto. apply transitive_closure_well_founded. apply H3.
apply H0. trivial.

intros D c t' Hy H. induction H.
intros. apply Acc_intro. intros. rename x into t'. rename y into u.
unfold symRel in H3.
assert (Type_WK D u c). eapply Well_kindedness_is_preserved_by_type_reduction; eauto.
destruct Hy. rename H5 into Hyp1. rename x into G.
destruct (Preservation_of_typing _ _ _ H4 _ Hyp1). destruct H5. destruct H5.
destruct H6. rename x into u'.
eapply H0. unfold symRel.
eapply Preservation_of_semantics.
apply H3. trivial. apply H5. trivial. trivial.
Qed.

(*******************************************************************)

Inductive TypeEnv_WK: TypeEnv -> Prop :=

  | WK_typeEnv_nil:
    (TypeEnv_WK nil)

  | WK_typeEnv_snoc:
    forall (D: TypeEnv) (Pi: TypeS),
      (TypeEnv_WK D) ->
      (Type_WK D Pi TypeSection) ->
      (TypeEnv_WK (snoc D Pi))
.

Lemma wellKinded_implies_trans_typeEnv_exists:
  forall (D: TypeEnv),
    (TypeEnv_WK D) ->
    exists D': LEnv,
      (trans_typeEnv D D')
.

Proof.
intros. induction H.
exists (nil (A := LType)). apply trans_typeEnv_nil.
destruct (wellKinded_implies_kind_of_exists _ _ _ H0). tauto. rename x into k.
destruct IHTypeEnv_WK. rename x into D'. exists (snoc D' (type_of_kind k)).
apply trans_typeEnv_snoc; trivial.
Qed.

(*******************************************************************)

Lemma Type_reduction_is_well_founded:
  forall (D: TypeEnv) (t: TypeS) (c: TypeC),
    (Type_WK D t c) ->
    (TypeEnv_WK D) ->
    (Acc (symRel RedType) t)
.

Proof.
intros.
destruct (wellKinded_implies_trans_typeEnv_exists _ H0).
destruct (Preservation_of_typing _ _ _ H _ H1). destruct H2.  destruct H2.
destruct H3.
eapply Preservation_of_semantics_cor1; eauto.
eapply WellTyped_implies_SN; eauto.
Qed.

End SetProgram.
