(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: Progress_proofs.v *)
(* Description: Progress *)

(** * Progress *)

Require Import List.
(* Require Import Arith. *)
(* Require Import Relations. *)

(* Require Import MatchCompNat. *)
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of TransitivityElimination_proofs.v *)
Require TransitivityElimination_proofs.
Module My_TransitivityElimination_proofs := TransitivityElimination_proofs.SetProgram(My_Program).
Import My_TransitivityElimination_proofs.

Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

Lemma evaluated_term_dec:
  forall (t: Term),
    (evaluated_term t) \/ ~(evaluated_term t)
.

Proof.
intros. induction t.
(* var *)
right. intro. inversion H.
(* select *)
right. intro. inversion H.
(* call *)
right. intro. inversion H.
(* new *)
destruct IHt. left. apply evaluated_new; trivial.
right. intro. apply H. inversion H0. trivial.
(* nil *)
left. apply evaluated_term_nil.
(* cons *)
destruct IHt1. destruct IHt2.
left. apply evaluated_term_cons; trivial.
right. intro. apply H0. inversion H1. trivial.
right. intro. apply H. inversion H0. trivial.
(* nil *)
left. apply evaluated_fieldTerm_nil.
(* cons *)
destruct IHt1. destruct IHt2.
left. apply evaluated_fieldTerm_cons; trivial.
right. intro. apply H0. inversion H1. trivial.
right. intro. apply H. inversion H0. trivial.
Qed.

(***************************************************************)

Lemma Progress_aux_1:
  forall (f: FieldSym) (fts: Term),
    (In f (fields_of_fieldTermList fts)) ->
    exists t: Term,
      (fieldTerm_lookup fts f t)
.

Proof.
intros.
induction fts.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. tauto.
simpl in H. rename fts1 into t0. rename fts2 into fts.
destruct H.
exists t0. rewrite H. apply fieldTerm_lookup_head.
destruct (IHfts2 H).
exists x.
apply fieldTerm_lookup_tail; trivial.
Qed.

(*****************************************************************)

Lemma evaluated_term_new_inversion:
  forall (D: TypeEnv) (G: TermEnv) (t: Term) (T: TypeS),
    (Term_WT D G t (TermI T)) ->
    (evaluated_term t) ->
    exists N: TypeS, exists fts: Term,
      t = (Term_new N fts) /\
      (Term_WT D G fts (FieldTermListI N))
.

Proof.
intros. inversion H.
rewrite <- H2 in H0. inversion H0.
rewrite <- H4 in H0. inversion H0.
rewrite <- H8 in H0. inversion H0.
exists N. exists fts. split. rewrite H1. trivial. trivial.
Qed.

(*******************************************************************)

Lemma KindS_at_is_a_partial_function:
  forall (ks k k': KindS) (i: nat),
    (KindS_at ks i k) ->
    (KindS_at ks i k') ->
    (k = k')
.

Proof.
intros ks k k' i H1 H2. generalize k' H2. clear k' H2. induction H1.
intros. inversion H2. trivial.
intros. inversion H2. apply IHKindS_at. trivial.
Qed.

Inductive SameCategory: TypeC -> TypeC -> Prop :=

  | SameCategory_TypeConstr:
    forall (k k': KindS),
      (SameCategory (TypeConstr k) (TypeConstr k'))

  | SameCategory_Type_:
    (SameCategory Type_ Type_)

  | SameCategory_TypeTuple:
    forall (k k': KindS),
      (SameCategory (TypeTuple k) (TypeTuple k'))

  | SameCategory_TypeParam:
    (SameCategory TypeParam TypeParam)

  | SameCategory_TypeSection:
    (SameCategory TypeSection TypeSection)

  | SameCategory_TypeOption:
    (SameCategory TypeOption TypeOption)

  | SameCategory_TypeParamList:
    (SameCategory TypeParamList TypeParamList)

  | SameCategory_TypeConstrList:
    forall (k k': KindS),
      (SameCategory (TypeConstrList k) (TypeConstrList k'))
.

Lemma Uniqueness_of_well_kindedness:
  forall (D: TypeEnv) (t: TypeS) (c1 c2: TypeC),
    (Type_WK D t c1) ->
    (Type_WK D t c2) ->
    (SameCategory c1 c2) ->
    (c1 = c2)
.

Proof.
intros. generalize c2 H0 H1. clear c2 H0 H1. induction H.
(* WK-proj *)
intros. inversion H1. apply f_equal with (f := TypeConstr).
eapply KindS_at_is_a_partial_function. apply H0. replace k with k0. trivial.
cut ((TypeTuple (Kind_tuple k)) = (TypeTuple (Kind_tuple k0))). intro.
injection H9. intro. symmetry. trivial.
eapply IHType_WK. trivial.
apply SameCategory_TypeTuple.
(* WK-class *)
intros. inversion H0. apply f_equal with (f := TypeConstr).
apply f_equal with (f := Kind_constr).
eapply kind_of_is_a_partial_function. apply H. trivial.
(* WK-fun *)
intros. inversion H3.
apply f_equal with (f := TypeConstr).
apply f_equal with (f := Kind_constr).
eapply kind_of_is_a_partial_function. apply H2.  trivial.
(* WK-apply *)
intros. inversion H1. trivial.
(* WK-var *)
intros. inversion H1.
apply f_equal with (f := TypeTuple).
eapply kind_of_is_a_partial_function. apply H0.  replace Pi with Pi0. trivial.
eapply decl_of_typeVar_is_a_partial_function. apply H4. trivial.
(* WK-tuple *)
intros. inversion H0.
apply f_equal with (f := TypeTuple).
apply f_equal with (f := Kind_tuple).
cut ((TypeConstrList k) = (TypeConstrList k0)). intro. inversion H6. trivial.
eapply IHType_WK. trivial. apply SameCategory_TypeConstrList.
(* WK-param *)
intros. inversion H1. trivial.
(* WK-section *)
intros. inversion H0. trivial.
(* WK-none *)
intros. inversion H0. trivial.
(* WK-some *)
intros. inversion H1. trivial.
(* WK-param-nil *)
intros. inversion H0. trivial. rewrite <- H3 in H1. inversion H1.
(* WK-param-cons *)
intros. inversion H1. trivial. rewrite <- H7 in H2. inversion H2.
(* WK-constr-nil *)
intros. inversion H0. rewrite <- H3 in H1. inversion H1. trivial.
(* WK-contsr-cons *)
intros. inversion H1. rewrite <- H7 in H2. inversion H2.
apply f_equal with (f := TypeConstrList).
apply f_equal2 with (f := Kind_cons).
cut ((TypeConstr k) = (TypeConstr k0)). intro. injection H9. tauto.
eapply IHType_WK1. trivial.
apply SameCategory_TypeConstr.
cut ((TypeConstrList ks) = (TypeConstrList ks0)). intro. injection H9. tauto.
eapply IHType_WK2. trivial.
apply SameCategory_TypeConstrList.
Qed.

(**********************************************************************)

Lemma Satisfaction_implies_same_kind:
  forall (D: TypeEnv) (R Pi: TypeS) (k: KindS),
    (Type_WK D R (TypeTuple k)) ->
    (SectionSat D R Pi) ->
    (kind_of Pi k)
.

Proof.
intros. inversion H0.
replace k with (Kind_tuple ks). apply Kind_of_section; trivial.
cut ((TypeTuple k) = (TypeTuple (Kind_tuple ks))). intro. injection H5. auto.
eapply Uniqueness_of_well_kindedness. apply H. apply H2.
apply SameCategory_TypeTuple.
Qed.

(******************************************************************)

Lemma Substitution_in_field_type:
  forall (D: TypeEnv) (f: FieldSym) (R: TypeS),
  let C := (fieldOwner f) in
  let T := (fieldType f) in
  (Type_WK D (Class_type C R) Type_) ->
  (Type_WK D (typeSubst_in_type T 0 R) Type_)
.

Proof.
intros.
inversion H. clear D0 K R0 H2 H0 H1 H.
inversion H3. clear D0 C0 k0 H0 H H1 H3.
destruct (field_declarations_are_well_formed f).
eapply Substitution_preserves_well_kindedness_snoc. apply H4.
apply H2.
replace (snoc D (classTypeParams C)) with (D ++ (Pi :: nil)).
apply Weakening_of_well_kindedness_by_the_left. trivial.
trivial.
Qed.

Lemma Substitution_in_method_type_params:
  forall (D: TypeEnv) (m: MethodSym) (R: TypeS),
  let C  := (methodOwner m) in
  let Pim := (methodTypeParams m) in
  (Type_WK D (Class_type C R) Type_) ->
  (Type_WK D (typeSubst_in_type Pim 0 R) TypeSection)
.

Proof.
intros.
inversion H. clear D0 K R0 H2 H0 H1 H.
inversion H3. clear D0 C0 k0 H0 H H1 H3.
destruct (method_declarations_are_well_formed m).
eapply Substitution_preserves_well_kindedness_snoc. eauto. eauto.
change (snoc D (classTypeParams C)) with (D ++ (PiC :: nil)).
apply Weakening_of_well_kindedness_by_the_left; trivial.
Qed.

Lemma Substitution_in_method_type:
  forall (D: TypeEnv) (m: MethodSym) (R R': TypeS),
  let C := (methodOwner m) in
  let T := (methodType m) in
  let Pi := (methodTypeParams m) in
  (Type_WK D (Class_type C R) Type_) ->
  (SectionSat D R' (typeSubst_in_type Pi 0 R)) ->
  (Type_WK D (typeSubst_in_type (typeSubst_in_type T 1 R) 0 R') Type_)
.

Proof.
intros.
inversion H. clear D0 K R0 H3 H1 H2 H.
inversion H4. clear D0 C0 k0 H1 H H2 H4.
inversion H0. clear H4 H0.
destruct (method_declarations_are_well_formed m). clear H4.
assert (kind_of (typeSubst_in_type Pim 0 R) (Kind_tuple ks)).
replace Pim with Pi. rewrite <- H. apply Kind_of_section. trivial. trivial.
eapply Substitution_preserves_well_kindedness_snoc. apply H2. apply H4.
replace (snoc D (typeSubst_in_type Pim 0 R)) with
        (D ++ (typeSubst_in_typeEnv (Pim :: nil) 0 R)).
replace 1 with (length (Pim :: nil)).
eapply Type_substitution_preserves_well_kindedness
with (D := D ++ PiC :: Pim :: nil).
apply H5. apply H3. 
apply Weakening_of_well_kindedness_by_the_left.
apply H6. trivial. trivial.
simpl. trivial.
Qed.

Lemma Substitution_in_method_param_types:
  forall (D: TypeEnv) (m: MethodSym) (R R': TypeS),
  let C := (methodOwner m) in
  let Ts := (methodParams m) in
  let Pi := (methodTypeParams m) in
  (Type_WK D (Class_type C R) Type_) ->
  (SectionSat D R' (typeSubst_in_type Pi 0 R)) ->
  (List_forall
     (fun T => (Type_WK D (typeSubst_in_type (typeSubst_in_type T 1 R) 0 R') Type_))
     Ts)
.

Proof.
intros.
inversion H. clear D0 K R0 H3 H1 H2 H.
inversion H4. clear D0 C0 k0 H1 H H2 H4.
inversion H0.
assert (kind_of (typeSubst_in_type Pi 0 R) (Kind_tuple ks)).
eapply Satisfaction_implies_same_kind; eauto.
destruct (method_declarations_are_well_formed m).
eapply List_forall_implies_forall with 
  (P1 := (fun U : TypeS => Type_WK D0 U Type_)).
intros. rename x into U.
eapply Substitution_preserves_well_kindedness_snoc. apply H2. apply H6.
replace (snoc D (typeSubst_in_type Pi 0 R)) with
        (D ++ (typeSubst_in_typeEnv (Pi :: nil) 0 R)).
replace 1 with (length (Pi :: nil)).
eapply Type_substitution_preserves_well_kindedness
with (D := D ++ PiC :: Pi :: nil).
apply H5. apply H3.
apply Weakening_of_well_kindedness_by_the_left.
apply H10. trivial. trivial.
simpl. trivial. trivial.
Qed.

Lemma Substitution_in_method_param_types_2:
  forall (D: TypeEnv) (m: MethodSym) (R: TypeS),
  let C := (methodOwner m) in
  let Ts := (methodParams m) in
  let Pim := (methodTypeParams m) in
  (Type_WK D (Class_type C R) Type_) ->
  (List_forall
     (fun T => (Type_WK (D ++ (typeSubst_in_type Pim 0 R) :: nil)
                        (typeSubst_in_type T 1 R) Type_))
     Ts)
.

Proof.
intros.
inversion H. clear D0 K R0 H2 H0 H1 H.
inversion H3. clear D0 C0 k0 H0 H H1 H3.
(* *)
destruct (method_declarations_are_well_formed m).
(* *)
eapply List_forall_implies_forall with 
  (P1 := (fun U : TypeS => Type_WK D0 U Type_)).
intros. rename x into U. unfold D0 in H3.
(* *)
assert (Type_WK (D ++ (PiC :: (Pim :: nil))) U Type_).
apply Weakening_of_well_kindedness_by_the_left. trivial.
clear H3.
(* *)
change (typeSubst_in_type Pim 0 R :: nil) with
       (typeSubst_in_typeEnv (Pim :: nil) 0 R).
change 1 with (length (Pim :: nil)).
eapply Type_substitution_preserves_well_kindedness; eauto.
(* *)
trivial.
Qed.

Lemma Substitution_in_method_type_2:
  forall (D: TypeEnv) (m: MethodSym) (R: TypeS),
  let C := (methodOwner m) in
  let T := (methodType m) in
  let Pim := (methodTypeParams m) in
  (Type_WK D (Class_type C R) Type_) ->
  (Type_WK (D ++ (typeSubst_in_type Pim 0 R) :: nil)
           (typeSubst_in_type T 1 R) Type_)
.

Proof.
intros.
inversion H. clear D0 K R0 H2 H0 H1 H.
inversion H3. clear D0 C0 k0 H0 H H1 H3.
(* *)
destruct (method_declarations_are_well_formed m).
(* *)
assert (Type_WK (D ++ (PiC :: (Pim :: nil))) T Type_).
apply Weakening_of_well_kindedness_by_the_left. trivial.
(* *)
change (typeSubst_in_type Pim 0 R :: nil) with
       (typeSubst_in_typeEnv (Pim :: nil) 0 R).
change 1 with (length (Pim :: nil)).
eapply Type_substitution_preserves_well_kindedness; eauto.
Qed.

(******************************************************************)

Definition Info_WK (D: TypeEnv) (info: TermInfo): Prop :=
  match info with
    | (TermI T)          => (Type_WK D T Type_)
    | (TermListI Ts)     => List_forall (fun T => (Type_WK D T Type_)) Ts
    | (FieldTermListI N) => True
  end.

Theorem Typing_returns_well_kinded_types:
  forall (D: TypeEnv) (G: TermEnv) (t: Term) (info: TermInfo),
    (Term_WT D G t info) ->
    (List_forall (fun T => (Type_WK D T Type_)) G) ->
    (Info_WK D info)
.

Proof.
intros D G t info H. induction H.
(* T_var *)
intros. simpl.
pattern T. eapply List_forall_and_List_from. apply H0. apply H.
(* T_select *)
intros. simpl.
unfold T.
apply Substitution_in_field_type. trivial.
(* T_call *)
intros. simpl.
unfold T.
apply Substitution_in_method_type. trivial. trivial.
(* T_new *)
intros. simpl. trivial.
(* T_nil *)
intros. simpl. apply List_forall_nil.
(* T_cons *)
intros. simpl. apply List_forall_cons; tauto.
(* T_fieldTerm_nil *)
intros. simpl. tauto.
(* T_fieldTerm_cons *)
intros. simpl. tauto.
Qed.

Lemma Typing_returns_well_kinded_types_cor1:
  forall (D: TypeEnv) (G: TermEnv) (t: Term) (T: TypeS),
    (Term_WT D G t (TermI T)) ->
    (List_forall (fun T => (Type_WK D T Type_)) G) ->
    (Type_WK D T Type_)
.

Proof.
intros. fold (Info_WK D (TermI T)).
eapply Typing_returns_well_kinded_types; eauto.
Qed.

(*****************************************************************)

(** Progress **)
Theorem Progress:
  forall (t: Term) (info: TermInfo),
    (Term_WT nil nil t info) ->
    ~(evaluated_term t) ->
    exists u: Term,
      (RedTerm t u)
.

Proof.
cut (
  forall (D: TypeEnv) (G: TermEnv) (t: Term) (info: TermInfo),
    (Term_WT D G t info) ->
    D = nil -> G = nil ->
    ~(evaluated_term t) ->
    exists u: Term,
      (RedTerm t u)
).
intros. eapply H; eauto.

intros D G t info H. induction H.

(* T_var *)
intros.
rewrite H1 in H. inversion H.
absurd ((snoc xs x)= nil). apply snoc_ind_aux_3. trivial.
absurd ((snoc xs x)= nil). apply snoc_ind_aux_3. trivial.

(* T_select *)
intros.
destruct (evaluated_term_dec t).
(* t is a value *)
destruct (evaluated_term_new_inversion _ _ _ _ H H5). destruct H6. destruct H6.
rename x into N. rename x0 into fts.
rewrite H6 in H. inversion H. clear fts0 H9.
rewrite <- H10 in H0. rewrite H2 in H0. unfold N0 in H0.
(* *)
assert (Subclass C0 C). eapply Subtyping_is_consistent_with_subclassing.
rewrite H2 in H11. unfold N0 in H11. apply H11.
rewrite H2 in H1. apply H1. trivial.
(* *)
inversion H14. clear C1 fs H16 H17 H18. unfold IsComplete_fields in H15.
assert (In f (fields_of_fieldTermList fts)); eauto.
destruct (Progress_aux_1 _ _ H16).
exists x.
rewrite H6.
apply R_select_redex. trivial.
(* t is not a value *)
destruct (IHTerm_WT H2 H3 H5). rename x into t'.
exists (Term_select t' f). apply R_select; trivial.

(* T_call *)
intros.
destruct (evaluated_term_dec t).
(* t is a value *)
destruct (evaluated_term_dec ts).
(* ts are values *)
clear IHTerm_WT1. clear IHTerm_WT2.
rewrite H6 in H. rewrite H6 in H0. rewrite H6 in H1. rewrite H6 in H2.
rewrite H6 in H3. rewrite H6 in H4. rewrite H6 in H5. clear D H6.
rewrite H7 in H. rewrite H7 in H1. clear G H7.
(* *)
destruct (evaluated_term_new_inversion _ _ _ _ H H9). destruct H6. destruct H6.
rename x into N. rename x0 into fts.
rewrite H6 in H. inversion H. clear fts0 H12.
rewrite <- H13 in H2. unfold N0 in H2.
(* *)
assert (Subclass C0 C). eapply Subtyping_is_consistent_with_subclassing.
unfold N0 in H14. apply H14. apply H3. trivial.
(* *)
destruct H17.  clear fs H17. unfold IsComplete_methods in H18.
destruct (H18 _ H12). clear H18. destruct H17. destruct H17.
rename x into C1. rename x0 into u.
destruct (Subclass_implies_class_expansion _ _ R0 H17).
rename x into R1.
exists (
  let u' := (typeSubst_in_term (typeSubst_in_term u 1 R1) 0 R) in
  (polyTermSubst_in_term u' 0 ((Term_new N fts) :: (list_of_termList ts)))
). rewrite H6. 
eapply R_call_redex.
rewrite <- H11. unfold N0. apply H19. trivial.
(* ts are not values *)
destruct (IHTerm_WT2 H6 H7 H10). rename x into ts'.
exists (Term_call t m R ts'). apply R_call_right; trivial.
(* t is not a value *)
destruct (IHTerm_WT1 H6 H7 H9). rename x into t'.
exists (Term_call t' m R ts). apply R_call_left; trivial.

(* T_new *)
intros.
destruct (evaluated_term_dec fts).
(* fts are evaluated *)
destruct H5. apply evaluated_new; trivial.
(* fts are not evaluated *)
destruct (IHTerm_WT H3 H4 H6). rename x into fts'.
exists (Term_new N fts'). apply R_new; trivial.

(* T_nil *)
intros.
destruct H1. apply evaluated_term_nil.

(* T_cons *)
intros.
destruct (evaluated_term_dec t).
(* t is evaluated *)
destruct (evaluated_term_dec ts).
(* ts is evaluated *)
destruct H3. apply evaluated_term_cons; trivial.
(* ts is not evaluated *)
destruct (IHTerm_WT2 H1 H2 H5). rename x into ts'.
exists (Term_cons t ts'). apply R_cons_right; trivial.
(* t is not evaluated *)
destruct (IHTerm_WT1 H1 H2 H4). rename x into t'.
exists (Term_cons t' ts). apply R_cons_left; trivial.

(* T_fieldTerm_nil *)
intros.
destruct H1. apply evaluated_fieldTerm_nil.

(* T_fieldTerm_cons *)
intros.
destruct (evaluated_term_dec t).
(* t is evaluated *)
destruct (evaluated_term_dec fts).
(* fts is evaluated *)
destruct H7. apply evaluated_fieldTerm_cons; trivial.
(* fts is not evaluated *)
destruct (IHTerm_WT2 H5 H6 H9). rename x into fts'.
exists (FieldTerm_cons f t fts'). apply R_fieldTerm_cons_right; trivial.
(* t is not evaluated *)
destruct (IHTerm_WT1 H5 H6 H8). rename x into t'.
exists (FieldTerm_cons f t' fts). apply R_fieldTerm_cons_left; trivial.
Qed.

End SetProgram.
