(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: TransitivityElimination_proofs.v *)
(* Description: Elimination of the transitivity rule *)

(** * Elimination of the transitivity rule *)

Require Import List.
(* Require Import Arith. *)
(* Require Import Relations. *)

(* Require Import MatchCompNat. *)
Require Import Util.
Require Import WellFoundation.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of Normalization_proofs.v *)
Require Normalization_proofs.
Module My_Normalization_proofs := Normalization_proofs.SetProgram(My_Program).
Import My_Normalization_proofs.

Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

(** Relation: Subtyping (w)ithout (t)ransivity (r)ule **)
Inductive SubtypingWTR: TypeEnv -> TypeS -> TypeS -> Prop :=

  | SWTR_Refl:
    forall (D: TypeEnv) (T: TypeS),
      (SubtypingWTR D T T)

  | SWTR_Var:
    forall (D: TypeEnv) (X: TypeVar) (i: nat) (R Ps Pi N N': TypeS),
      (decl_of_typeVar D X (Type_section Ps)) ->
      (TypeS_at Ps i (Type_param Pi (Type_some N))) ->
      (SubtypingWTR D (typeSubst_in_type N 0 R) N') ->
      (SubtypingWTR D (Var_type X i R) N')

  | SWTR_Class:
    forall (D: TypeEnv) (C: ClassSym) (R: TypeS) (N N': TypeS),
      (classSuper C) = (Some N) ->
      (SubtypingWTR D (typeSubst_in_type N 0 R) N') ->
      (SubtypingWTR D (Class_type C R) N')

  | SWTR_Beta_Left:
    forall (D: TypeEnv) (T T' U: TypeS),
      (RedType T T') ->
      (SubtypingWTR D T' U) ->
      (SubtypingWTR D T U)

  | SWTR_Beta_Right:
    forall (D: TypeEnv) (T U' U: TypeS),
      (RedType U U') ->
      (SubtypingWTR D T U') ->
      (SubtypingWTR D T U)
.

(** Relation: Type expansion **)
Inductive TypeExp: TypeEnv -> TypeS -> TypeS -> Prop :=

(*   | TE_Var: *)
(*     forall (D: TypeEnv) (X: TypeVar) (i: nat) (R Ps Pi N: TypeS), *)
(*       (decl_of_typeVar D X (Type_section Ps)) -> *)
(*       (TypeS_at Ps i (Type_param Pi (Type_some N))) -> *)
(*       (TypeExp D (Var_type X i R) (typeSubst_in_type N 0 R)) *)

  | TE_Class:
    forall (D: TypeEnv) (C: ClassSym) (R: TypeS) (N: TypeS),
      (classSuper C) = (Some N) ->
      (TypeExp D (Class_type C R) (typeSubst_in_type N 0 R))

  | TE_Beta:
    forall (D: TypeEnv) (T T': TypeS),
      (RedType T T') ->
      (TypeExp D T T')
.

(*******************************************************************)

Lemma SWTR_beta_right_star:
  forall (D : TypeEnv) (T U' U : TypeS),
    (SubtypingWTR D T U') -> 
    (RedTypeStar U U') ->
    (SubtypingWTR D T U)
.

Proof.
intros D T U' U H1 H2. induction H2.
(* Step *)
eapply SWTR_Beta_Right.  apply H. trivial.
(* Refl *)
trivial.
(* Trans *)
tauto.
Qed.

Lemma SWTR_beta_left_star:
  forall (D : TypeEnv) (T T' U : TypeS),
    (RedTypeStar T T') ->
    (SubtypingWTR D T' U) -> 
    (SubtypingWTR D T U)
.

Proof.
intros D T U' U H1 H2. induction H1.
(* Step *)
eapply SWTR_Beta_Left.  apply H. trivial.
(* Refl *)
trivial.
(* Trans *)
tauto.
Qed.

(** Subtyping is preserved by type reduction *)
Lemma Reduction_preserves_subtypingWTR:
  forall (D D': TypeEnv) (T U T' U': TypeS),
    (SubtypingWTR D T U) ->
    (RedTypeEnvStar D D') ->
    (RedTypeStar T T') ->
    (RedTypeStar U U') ->
    (SubtypingWTR D' T' U')
.

Proof.
intros D D' T U T' U' H. generalize D' T' U'. clear D' T' U'. induction H.
(* S-Refl *)
intros.
destruct (Type_reduction_is_confluent _ _ _ H0 H1). destruct H2. rename x into S.
eapply SWTR_beta_right_star.
eapply SWTR_beta_left_star. apply H2. apply SWTR_Refl. trivial.
(* S-Var *)
intros.
destruct (Reduction_of_variable_types_star _ _ _ _ H3). destruct H5.
rename x into R'. rewrite H5.
destruct (Declaration_of_type_variable_is_preserved_by_reduction _ _ _ _ H H2).
destruct H7. rename x into Ps'.
destruct (TypeS_at_is_preserved_by_reduction _ _ _ _ _ H0 H8). destruct H9.
destruct H9. destruct H10. 
eapply SWTR_Var. apply H7. apply H9.
eapply IHSubtypingWTR. trivial.
apply Reduction_and_substitution_star; trivial.
trivial.
(* S-Class *)
intros.
destruct (Reduction_of_class_types_star _ _ _ H2). destruct H4.
rename x into R'. rewrite H4.
eapply SWTR_Class. apply H.
eapply IHSubtypingWTR. trivial.
apply Reduction_in_substitution_star; trivial.
trivial.
(* S-Beta-Left *)
intros.
assert (RedTypeStar T T'). apply RedType_step; trivial.
destruct (Type_reduction_is_confluent _ _ _ H4 H2). destruct H5. rename x into S.
eapply SWTR_beta_left_star. apply H6. apply IHSubtypingWTR. trivial. trivial. trivial.
(* S-Beta-Right *)
intros.
assert (RedTypeStar U U'). apply RedType_step; trivial.
destruct (Type_reduction_is_confluent _ _ _ H4 H3). destruct H5. rename x into S.
eapply SWTR_beta_right_star. apply IHSubtypingWTR. trivial. trivial. apply H5. trivial.
Qed.

(******************************************************************************)

Lemma Class_type_expansion_preserves_well_kindedness:
  forall (D: TypeEnv) (C: ClassSym) (R N: TypeS),
  (Type_WK D (Class_type C R) Type_) ->
  (classSuper C) = (Some N) ->
  (Type_WK D (typeSubst_in_type N 0 R) Type_)    
.

Proof.
intros.
inversion H. clear D0 K R0 H3 H1 H2. inversion H4. clear D0 C0 k0 H2 H1 H3.
destruct (classes_are_well_kinded C). destruct (H2 _ H0); clear H2. destruct H7.
unfold Pi in H2.
eapply Substitution_preserves_well_kindedness_snoc. apply H5. apply H6.
unfold snoc. apply Weakening_of_well_kindedness_by_the_left. trivial.
Qed. 

Lemma classtype_expansion_preserves_well_kindedness_1:
  forall (D: TypeEnv) (C C': ClassSym) (R R': TypeS),
  (Type_WK D (Class_type C R) Type_) ->
  (classSuper C) = (Some (Class_type C' R')) ->
  (Type_WK D (Class_type C' (typeSubst_in_type R' 0 R)) Type_)    
.

Proof.
intros.
inversion H. clear D0 K R0 H3 H1 H2. inversion H4. clear D0 C0 k0 H2 H1 H3.
destruct (classes_are_well_kinded C). destruct (H2 _ H0); clear H2. destruct H7.
unfold Pi in H2.
replace (Class_type C' (typeSubst_in_type R' 0 R)) with 
  (typeSubst_in_type (Class_type C' R') 0 R).
eapply Substitution_preserves_well_kindedness_snoc. apply H5. apply H6.
unfold snoc. apply Weakening_of_well_kindedness_by_the_left. trivial.
trivial.
Qed. 

Lemma Type_expansion_is_well_founded_aux1:
  forall (D: TypeEnv) (C: ClassSym) (R: TypeS),
    (Type_WK D (Class_type C R) Type_) ->
    (TypeEnv_WK D) ->
    (Acc (symRel (TypeExp D)) (Class_type C R))
.

Proof.
cut (
  forall (D: TypeEnv),
    (TypeEnv_WK D) ->
    forall (C: ClassSym) (T: TypeS),
      (Type_WK D T Type_) ->
      forall (R: TypeS),
        T = (Class_type C R) ->      
        (Acc (symRel (TypeExp D)) (Class_type C R))
).
intros. eapply H; eauto.

intros D Hy C.
induction (Class_inheritance_is_well_founded C). clear H. rename x into C.
intros T H1.
assert (Acc (symRel RedType) T).
eapply Type_reduction_is_well_founded. apply H1. trivial. induction H. clear H.
intros. rewrite H in H1. rewrite H in H2. clear x H.
apply Acc_intro. intros. rename y into T. unfold symRel in H.
inversion H.
(* TE-Var *)
(* TE-Class *)
destruct (classes_are_well_kinded C). destruct (H9 _ H7).
destruct H10. destruct H10. rename x into C'. rename x0 into R'.
rewrite H10. simpl. rewrite H10 in H7. rewrite H10 in H11.
fold (Class_type C' (typeSubst_in_type R' 0 R)).
eapply H0 with (T := Class_type C' (typeSubst_in_type R' 0 R)).
red. eapply classInheritance_intro. apply H7.
eapply classtype_expansion_preserves_well_kindedness_1. apply H1. trivial. trivial.
(* TE-Red *)
clear D0 T0 T' H4 H5 H6.
destruct (Reduction_of_class_types _ _ _ H3). destruct H4. rename x into R'.
rewrite H4. eapply H2. unfold symRel. apply H3.
eapply Well_kindedness_is_preserved_by_type_reduction.
apply H1. trivial. trivial.
Qed.

(* Lemma Type_expansion_is_well_founded_aux2: *)
(*   forall (D: TypeEnv) (X: TypeVar) (i: nat) (R: TypeS), *)
(*     (Type_WK D (Var_type X i R) Type_) -> *)
(*     (Acc (symRel (TypeExp D)) (Var_type X i R)) *)
(* . *)

(* Proof. *)
(* cut ( *)
(*   forall (D: TypeEnv) (T: TypeS), *)
(*     (Type_WK D T Type_) -> *)
(*     forall (X: TypeVar) (i: nat) (R: TypeS), *)
(*       T = (Var_type X i R) -> *)
(*       (Acc (symRel (TypeExp D)) T) *)
(* ). *)
(* intros. eapply H. trivial. reflexivity. *)

(* intros D T H. *)
(* assert (Acc (symRel RedType) T). *)
(* eapply Type_reduction_is_well_founded. apply H. induction H0. *)
(* intros. rename x into T. *)
(* apply Acc_intro. intros. rename y into U. unfold symRel in H3. *)
(* rewrite H2 in H3. *)
(* inversion H3. *)
(* (\* TE-Var *\) *)
(* rewrite H2 in H. *)
(* (\* inversion H. clear D1 K R1 H13 H11 H12. *\) *)
(* apply AdmitIt. *)
(* (\* destruct (classes_are_well_kinded C). destruct (H9 _ H7). *\) *)
(* (\* destruct H10. destruct H10. rename x into C'. rename x0 into R'. *\) *)
(* (\* rewrite H10. simpl. rewrite H10 in H7. rewrite H10 in H11. *\) *)
(* (\* fold (Class_type C' (typeSubst_in_type R' 0 R)). *\) *)
(* (\* eapply H0 with (T := Class_type C' (typeSubst_in_type R' 0 R)). *\) *)
(* (\* red. eapply classInheritance_intro. apply H7. *\) *)
(* (\* eapply classtype_expansion_preserves_well_kindedness. apply H1. trivial. trivial. *\) *)
(* (\* TE-Class *\) *)
(* (\* TE-Red *\) *)
(* clear D0 T0 T' H5 H6 H7. *)
(* destruct (Reduction_of_variable_types _ _ _ _ H4). destruct H5. rename x into R'. *)
(* eapply H1. unfold symRel. rewrite H2. trivial. *)
(* eapply Well_kindedness_is_preserved_by_type_reduction. apply H. *)
(* rewrite H2. trivial. apply H5. *)
(* Qed. *)

Lemma Type_expansion_is_well_founded:
  forall (D: TypeEnv) (T: TypeS),
    (TypeEnv_WK D) ->
    (Type_WK D T Type_) ->
    (Acc (symRel (TypeExp D)) T)
.

Proof.
intros D T Hy H.
assert (Acc (symRel RedType) T).
eapply Type_reduction_is_well_founded; eauto.
induction H0. rename x into T.
apply Acc_intro. intros. rename y into U. unfold symRel in H2. destruct H2.
(* (\* TE-Var *\) *)
(* assert (Acc (symRel (TypeExp D)) (Var_type X i R)). *)
(* apply Type_expansion_is_well_founded_aux2. trivial. inversion H4. eapply H5. *)
(* unfold symRel. eapply TE_Var. apply H2. apply H3. *)
(* TE-Class *)
assert (Acc (symRel (TypeExp D)) (Class_type C R)).
apply Type_expansion_is_well_founded_aux1; trivial. inversion H3. eapply H4.
unfold symRel. eapply TE_Class.  trivial.
(* TE-Red *)
eapply H1. unfold symRel. trivial.
eapply Well_kindedness_is_preserved_by_type_reduction. apply H. trivial.
Qed.

(* Lemma Type_expansion_preserves_well_kindedness: *)
(*   forall (D: TypeEnv) (T U: TypeS), *)
(*     (Type_WK D T Type_) -> *)
(*     (TypeExp D T U) -> *)
(*     (Type_WK D U Type_) *)
(* . *)

(* Proof. *)
(* intros. *)
(* destruct H0. *)
(* (\* TE-Var *\) *)
(* apply AdmitIt. *)
(* (\* TE-Class *\) *)
(* eapply Class_type_expansion_preserves_well_kindedness. apply H. trivial. *)
(* (\* TE-Red *\) *)
(* eapply Well_kindedness_is_preserved_by_type_reduction. apply H. trivial. *)
(* Qed. *)

(*********************************************************************)

Lemma SWTR_Trans_aux:
    forall (T S U: TypeS),
    (Acc (symRel (TypeExp nil)) T) ->
    (Acc (symRel (TypeExp nil)) S) ->
    (Acc (symRel (TypeExp nil)) U) ->
    (SubtypingWTR nil T S) ->
    (SubtypingWTR nil S U) ->
    (SubtypingWTR nil T U)
.

Proof.
cut (
  forall (Ts: (list TypeS)),
    (Acc (RList TypeS (symRel (TypeExp nil))) Ts) ->
    forall (T S U: TypeS),
      Ts = (T :: S :: U :: nil) ->
      (SubtypingWTR nil T S) ->
      (SubtypingWTR nil S U) ->
      (SubtypingWTR nil T U)
).
intros. eapply H.
apply sequence_extension_preserves_accessibility.
apply List_forall3. apply H0. apply H1. apply H2. reflexivity. trivial. trivial.

intros Ts H. induction H.
intros. rename x into Ts.
inversion H2.
(* SWRT-Refl,? *)
trivial.
(* SWRT-Var,? *)
inversion H4. inversion H13.
absurd ((snoc D1 Pi1) = nil). apply snoc_ind_aux_3. trivial.
absurd ((snoc D1 Pi1) = nil). apply snoc_ind_aux_3. trivial.
(* SWRT-Class,? *)
eapply SWTR_Class. apply H4. eapply H0. rewrite H1.
apply RTriple1. unfold symRel. rewrite <- H7. apply TE_Class. apply H4.
reflexivity. trivial. trivial.
(* SWRT-Beta-Left *)
eapply SWTR_Beta_Left. apply H4. eapply H0. rewrite H1.
apply RTriple1. unfold symRel. apply TE_Beta. apply H4.
reflexivity. trivial. trivial.
(* SWRT-Beta-Right *)
rename U' into T'.
inversion H3.
(* - SWRT-Beta-Right, SWRT-Refl *)
rewrite <- H11. trivial.
(* - SWRT-Beta-Right, SWRT-Var *)
inversion H9. inversion H18.
absurd ((snoc D2 Pi1) = nil). apply snoc_ind_aux_3. trivial.
absurd ((snoc D2 Pi1) = nil). apply snoc_ind_aux_3. trivial.
(* - SWRT-Beta-Right, SWRT-Class *)
rewrite <- H12 in H4.
destruct (Reduction_of_class_types _ _ _ H4). destruct H14.
eapply H0. rewrite H1. apply RTriple2. rewrite <- H12. unfold symRel.
apply TE_Beta. apply H4. reflexivity. trivial. rewrite H14. rename x into R'.
eapply SWTR_Class. apply H9.
eapply Reduction_preserves_subtypingWTR with (D := (nil (A := TypeS)))
      (T := (typeSubst_in_type N 0 R)) (U := U).
trivial.
apply RedTypeEnv_refl.
apply Reduction_in_substitution. trivial. apply RedType_refl.
(* - SWRT-Beta-Right, SWRT-Beta-Left *)
rename T'0 into U'.
assert (RedTypeStar S T'). apply RedType_step. trivial.
assert (RedTypeStar S U'). apply RedType_step. trivial.
destruct (Type_reduction_is_confluent _ _ _ H14 H15). destruct H16. rename x into S'.
eapply H0. rewrite H1. apply RTriple2. unfold symRel. apply TE_Beta. apply H4. reflexivity.
trivial.
eapply SWTR_beta_left_star. apply H16.
eapply Reduction_preserves_subtypingWTR with (D := (nil (A := TypeS)))
      (T := U') (U := U).
trivial.
apply RedTypeEnv_refl. trivial. apply RedType_refl.
(* - SWRT-Beta-Right (actually ?), SWRT-Beta-Right *)
eapply SWTR_Beta_Right. apply H9. eapply H0. rewrite H1.
apply RTriple3. unfold symRel. apply TE_Beta. apply H9.
reflexivity. trivial. trivial.
Qed.

Lemma SWTR_Trans:
  forall (T S U: TypeS),
    (Type_WK nil T Type_) ->
    (Type_WK nil S Type_) ->
    (Type_WK nil U Type_) ->
    (SubtypingWTR nil T S) ->
    (SubtypingWTR nil S U) ->
    (SubtypingWTR nil T U)
.

Proof.
assert (TypeEnv_WK nil). apply WK_typeEnv_nil. rename H into Hy.
intros.
eapply SWTR_Trans_aux with (S := S).
apply Type_expansion_is_well_founded; trivial.
apply Type_expansion_is_well_founded; trivial.
apply Type_expansion_is_well_founded; trivial.
trivial. trivial.
Qed.

Lemma TransitivityElimination_aux:
  forall (T U: TypeS),
    (Acc (symRel (TypeExp nil)) T) ->
    (Acc (symRel (TypeExp nil)) U) ->
    (Subtyping nil T U) ->
    (SubtypingWTR nil T U)
.

Proof.
cut (
  forall (D: TypeEnv) (T U: TypeS),
    (Subtyping D T U) ->
    D = nil ->
    (Acc (symRel (TypeExp nil)) T) ->
    (Acc (symRel (TypeExp nil)) U) ->
    (SubtypingWTR nil T U)
).
intros. eapply H. apply H2. trivial. trivial. trivial.
intros D T U H. induction H.
(* S-Refl *)
intros. apply SWTR_Refl.
(* S-Trans *)
intros. rewrite H2 in H. rewrite H2 in H1.
assert (Acc (symRel (TypeExp nil)) S).
apply Type_expansion_is_well_founded. apply WK_typeEnv_nil. trivial.
apply SWTR_Trans_aux with (S := S). trivial. trivial. tauto. tauto. tauto.
(* S-Var*)
intros.
rewrite H2 in H. inversion H. inversion H8.
absurd ((snoc D1 Pi1) = nil). apply snoc_ind_aux_3. trivial.
absurd ((snoc D1 Pi1) = nil). apply snoc_ind_aux_3. trivial.
(* intros. rewrite H2 in H. eapply SWTR_Var. apply H. apply H0. apply IHSubtyping. *)
(* trivial. inversion H3. apply H5. unfold symRel. eapply TE_Var. apply H. apply H0. *)
(* trivial. *)
(* S-Class *)
intros. eapply SWTR_Class. apply H. apply IHSubtyping. trivial.
inversion H2. apply H4. unfold symRel. eapply TE_Class. trivial. trivial.
(* S-Beta-Left *)
intros. eapply SWTR_Beta_Left. apply H. apply IHSubtyping. trivial.
inversion H2. apply H4. unfold symRel. apply TE_Beta. trivial. trivial.
(* S-Beta-Right *)
intros. eapply SWTR_Beta_Right. apply H. apply IHSubtyping. trivial. trivial.
inversion H3. apply H4. unfold symRel. apply TE_Beta. trivial.
Qed.

Theorem TransitivityElimination:
  forall (T U: TypeS),
    (Type_WK nil T Type_) ->
    (Type_WK nil U Type_) ->
    (Subtyping nil T U) ->
    (SubtypingWTR nil T U)
.

Proof.
assert (TypeEnv_WK nil). apply WK_typeEnv_nil. rename H into Hy.
intros. apply TransitivityElimination_aux.
apply Type_expansion_is_well_founded; trivial.
apply Type_expansion_is_well_founded; trivial.
trivial.
Qed.

Lemma SubtypingWTR_is_consistent_with_subclassing:
  forall (D: TypeEnv) (C C': ClassSym) (R R': TypeS),
    (SubtypingWTR D (Class_type C R) (Class_type C' R')) ->
    (Subclass C C')
.

Proof.
cut (
  forall (D: TypeEnv) (T T': TypeS),
    (SubtypingWTR D T T') ->
    forall (C C': ClassSym) (R R': TypeS),
      T = (Class_type C R) ->
      T' = (Class_type C' R') ->
      (Subclass C C')
). intros. eapply H. apply H0. reflexivity. reflexivity.
intros D T T' H. induction H.
(* SWTR-Refl *)
intros. rewrite H in H0. injection H0; intros.
replace C' with C. apply SC_Refl.
(* SWTR-Var *)
intros. inversion H2.
(* SWTR-Class *)
intros. injection H1; intros. rewrite <- H4.
destruct (classes_are_well_kinded C). destruct (H6 _ H).
destruct H7. destruct H7. rewrite H7 in H.
eapply SC_Extends.  apply H.
eapply IHSubtypingWTR. rewrite H7. simpl. unfold Class_type. reflexivity.
apply H2.
(* SWTR-Beta-Left *)
intros. rewrite H1 in H.
destruct (Reduction_of_class_types _ _ _ H). destruct H3.
eapply IHSubtypingWTR. apply H3. apply H2.
(* SWTR-Beta-Right *)
intros. rewrite H2 in H.
destruct (Reduction_of_class_types _ _ _ H). destruct H3.
eapply IHSubtypingWTR. apply H1. apply H3.
Qed.

Theorem Subtyping_is_consistent_with_subclassing:
  forall (C C': ClassSym) (R R': TypeS),
    (Type_WK nil (Class_type C R) Type_) ->
    (Type_WK nil (Class_type C' R') Type_) ->
    (Subtyping nil (Class_type C R) (Class_type C' R')) ->
    (Subclass C C')
.

Proof.
intros.
eapply SubtypingWTR_is_consistent_with_subclassing.
eapply TransitivityElimination. apply H. apply H0. trivial.
Qed.

End SetProgram.
