(* Author: Vincent Cremet *)
(* Date: June 2007 *) 
(* File: TypeSubstitutionInTyping.v *)
(* Description: Type substitution in typing *)

(** * Type substitution in typing *)

Require Import List.
(* Require Import Arith. *)
(* Require Import Relations. *)

(* Require Import MatchCompNat. *)
Require Import Util.

Require Syntax.
Module SetProgram(My_Program: Syntax.Program).
Import My_Program.

(* Import definitions of BoundCompliantSubstitutions_proofs.v *)
Require BoundCompliantSubstitutions_proofs.
Module My_BoundCompliantSubstitutions_proofs := BoundCompliantSubstitutions_proofs.SetProgram(My_Program).
Import My_BoundCompliantSubstitutions_proofs.

Import My_Progress_proofs.
Import My_TransitivityElimination_proofs.
Import My_Normalization_proofs.
Import My_Subtyping_proofs.
Import My_Typing_proofs.
Import My_Confluence.
Import My_Auxiliary_proofs.
Import My_Substitutions_proofs.
Import My_Binders_proofs.
Import My_Semantics.
Import My_WellFormedness.
Import My_Typing.
Import My_Auxiliary.
Import My_Substitutions.
Import My_Binders.

(********************************************************************)
(********************************************************************)

Definition typeSubst_in_termInfo (info: TermInfo) (k: nat) (R: TypeS): TermInfo :=
  match info with
    | (TermI T) => (TermI (typeSubst_in_type T k R))
    | (TermListI Ts) => (TermListI (map (fun T => (typeSubst_in_type T k R)) Ts))
    | (FieldTermListI N) => (FieldTermListI (typeSubst_in_type N k R))
  end
.

(**********************************************************************)

Definition TermEnv_WK (D: TypeEnv) (G: TermEnv): Prop :=
  List_forall (fun T => (Type_WK D T Type_)) G
.

(**********************************************************************)

Lemma Type_substitution_preserves_fields_of_fieldTermList:
  forall (fts: Term) (k: nat) (R: TypeS),
    (fields_of_fieldTermList (typeSubst_in_term fts k R)) =
    (fields_of_fieldTermList fts)
.

Proof.
intros. induction fts.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. trivial.
simpl. rewrite IHfts2. trivial.
Qed.

(**********************************************************************)

Lemma Permuting_type_substitutions_in_lists:
  forall (R R' : TypeS) (k : nat) (ts : (list TypeS)) (n : nat),
    (typeSubst_in_typeList (typeSubst_in_typeList ts n R) (n + k) R') =
    (typeSubst_in_typeList (typeSubst_in_typeList ts (S (n + k)) R')
                           n (typeSubst_in_type R k R'))
.

Proof.
intros. induction ts.
simpl. trivial.
simpl. rewrite IHts. rewrite Permuting_type_substitutions. trivial.
Qed.

Lemma Permuting_type_substitutions_in_lists_cor1:
  forall (R R' : TypeS) (k : nat) (ts : (list TypeS)),
    (typeSubst_in_typeList (typeSubst_in_typeList ts 0 R) k R') =
    (typeSubst_in_typeList (typeSubst_in_typeList ts (S k) R')
                           0 (typeSubst_in_type R k R'))
.

Proof.
intros. induction ts.
simpl. trivial.
simpl. rewrite IHts. rewrite Permuting_type_substitutions_cor1. trivial.
Qed.

Lemma Permuting_type_substitutions_in_lists_cor2:
  forall (R R' : TypeS) (k : nat) (ts : (list TypeS)),
    (typeSubst_in_typeList (typeSubst_in_typeList ts 1 R) (S k) R') =
    (typeSubst_in_typeList (typeSubst_in_typeList ts (S (S k)) R')
                           1 (typeSubst_in_type R k R'))
.

Proof.
intros. induction ts.
simpl. trivial.
simpl. rewrite IHts. rewrite Permuting_type_substitutions_cor2. trivial.
Qed.

(**********************************************************************)

Theorem Type_substitution_preserves_typing:
  forall (D1 D2: TypeEnv) (G: TermEnv) (Pi R: TypeS) (k: KindS) (t: Term) (info: TermInfo),
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK (D1 ++ (Pi :: D2))) ->
    (TermEnv_WK (D1 ++ (Pi :: D2)) G) ->
    (Term_WT (D1 ++ (Pi :: D2)) G t info) ->
    (Term_WT (D1 ++ (typeSubst_in_typeEnv D2 0 R))
             (map (fun T => (typeSubst_in_type T (length D2) R)) G)
             (typeSubst_in_term t (length D2) R)
             (typeSubst_in_termInfo info (length D2) R))
.

Proof.
cut (
  forall (D1 D2 D: TypeEnv) (G: TermEnv) (Pi R: TypeS)
         (k: KindS) (t: Term) (info: TermInfo),
    (Type_WK D1 R (TypeTuple k)) ->
    (SectionSat D1 R Pi) ->
    (TypeEnv_WK (D1 ++ (Pi :: D2))) ->
    (TermEnv_WK (D1 ++ (Pi :: D2)) G) ->
    (Term_WT D G t info) ->
    D = (D1 ++ (Pi :: D2)) ->
    (Term_WT (D1 ++ (typeSubst_in_typeEnv D2 0 R))
             (map (fun T => (typeSubst_in_type T (length D2) R)) G)
             (typeSubst_in_term t (length D2) R)
             (typeSubst_in_termInfo info (length D2) R))
).
intros. eapply H; eauto.

intros. generalize H4. clear H4. induction H3.

(* T-var *)
intros. simpl. apply T_Var.
change (typeSubst_in_type T (length D2) R) with
  ((fun U => (typeSubst_in_type U (length D2) R)) T).
apply List_from_of_map. trivial.

(* T-select *)
rename S into T_t.
intros. simpl.
change (typeSubst_in_type (typeSubst_in_type T 0 R0) (length D2) R) with
       (typeSubst_in_type (typeSubst_in_type T 0 R0) (0 + (length D2)) R).
rewrite Permuting_type_substitutions. simpl. unfold T.
replace (typeSubst_in_type (fieldType f) (S (length D2)) R) with (fieldType f).
eapply T_Select. simpl in IHTerm_WT. apply IHTerm_WT; trivial.
fold C.
rewrite <- typeSubst_of_class_type.
eapply Type_substitution_preserves_subtyping. eauto. eauto. eauto.
rewrite <- H6. trivial. rewrite H6 in H3.
eapply Typing_returns_well_kinded_types_cor1. apply H3.
apply H2. rewrite <- H6. trivial.
rewrite <- typeSubst_of_class_type. fold C. rewrite H6 in H5.
eapply Type_substitution_preserves_well_kindedness.
eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
change (S (length D2)) with ((length ((classTypeParams C) :: nil)) + (length D2)).
symmetry. eapply Substitution_in_closed_type_aux.
destruct (field_declarations_are_well_formed f).
apply H7.

(* T-call *)
rename S into T_t. rename Ss into T_ts.
intros. simpl.
replace (typeSubst_in_type (typeSubst_in_type (typeSubst_in_type T 1 R') 0 R0) 
                           (length D2) R) with
        (typeSubst_in_type (typeSubst_in_type T 1 (typeSubst_in_type R' (length D2) R)) 
                           0 (typeSubst_in_type R0 (length D2) R)).
unfold T. eapply T_Call.
(* *)
simpl in IHTerm_WT1. apply IHTerm_WT1; trivial.
(* *)
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
(* *)
simpl in IHTerm_WT2. apply IHTerm_WT2; trivial.
(* *)
clear IHTerm_WT1. clear IHTerm_WT2.
fold C. rewrite <- typeSubst_of_class_type.
eapply Type_substitution_preserves_subtyping. eauto. eauto. eauto.
rewrite <- H8. trivial. rewrite H8 in H3_.
eapply Typing_returns_well_kinded_types_cor1. apply H3_.
trivial.
rewrite <- H8. trivial.
(* *)
clear IHTerm_WT1. clear IHTerm_WT2.
fold C. rewrite <- typeSubst_of_class_type.
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
(* *)
clear IHTerm_WT1. clear IHTerm_WT2.
fold Pi0.
replace Pi0 with (typeSubst_in_type Pi0 (S (0 + length D2)) R).
rewrite <- Permuting_type_substitutions. simpl.
eapply Type_substitution_preserves_satisfaction. eauto. eauto. trivial.
rewrite <- H8. trivial.
(* *)
unfold Pi0.
eapply Substitution_in_method_type_params.
rewrite <- H8. trivial.
(* *)
simpl.
change (S (length D2)) with ((length ((classTypeParams C) :: nil)) + (length D2)).
eapply Substitution_in_closed_type_aux.
destruct (method_declarations_are_well_formed m). apply H9.
(* *)
clear IHTerm_WT1. clear IHTerm_WT2.
fold Ts.
replace (typeSubst_in_typeList
          (typeSubst_in_typeList Ts 1 (typeSubst_in_type R' (length D2) R))
          0 (typeSubst_in_type R0 (length D2) R)) with
       (typeSubst_in_typeList
          (typeSubst_in_typeList (typeSubst_in_typeList Ts 1 R') 0 R0)
          (length D2) R).
unfold Pointwise_subtyping.
unfold typeSubst_in_typeList.
eapply Forall2_of_map with (P1 := fun T U =>
  ((Type_WK D T Type_) /\
   (Type_WK D U Type_)) /\
  (Subtyping D T U)).
intros. destruct H9. destruct H9.
eapply Type_substitution_preserves_subtyping. eauto. eauto. trivial.
rewrite <- H8; trivial. rewrite <- H8; trivial. rewrite <- H8; trivial.
apply Forall2_conj with
  (P1 := fun T U => (Type_WK D T Type_ /\ Type_WK D U Type_))
  (P2 := fun T U => (Subtyping D T U)).
apply List_forall_implies_Forall2 with
  (P1 := fun T => (Type_WK D T Type_))
  (P2 := fun T => (Type_WK D T Type_))
.
cut (Info_WK D (TermListI T_ts)). tauto.
eapply Typing_returns_well_kinded_types. eauto.
rewrite H8. trivial.
rewrite length_of_map. rewrite length_of_map.
(* *)
assert ((length T_ts) =
        (length (typeSubst_in_typeList (typeSubst_in_typeList Ts 1 R') 0 R0))).
eapply Forall2_implies_same_length. apply H7.
unfold typeSubst_in_typeList in H9. rewrite length_of_map in H9.
unfold typeSubst_in_typeList in H9. rewrite length_of_map in H9. trivial.
(* *)
apply List_forall_of_map. apply List_forall_of_map.
unfold Ts. apply Substitution_in_method_param_types; eauto.
(* *)
apply H7.
(* *)
rewrite Permuting_type_substitutions_in_lists_cor1.
rewrite Permuting_type_substitutions_in_lists_cor2.
replace (typeSubst_in_typeList Ts (S (S (length D2))) R) with Ts. trivial.
change (S (S (length D2))) with
       ((length ((classTypeParams C) :: Pi0 :: nil)) + (length D2)).
symmetry.
unfold typeSubst_in_typeList.
apply List_forall_equal_maps_cor1.
apply List_forall_implies_forall with
  (P1 := fun T => (Type_WK ((classTypeParams C) :: Pi0 :: nil) T Type_)).
intros. rename x into U.
eapply Substitution_in_closed_type_aux. eauto.
destruct (method_declarations_are_well_formed m). trivial.
(* *)
clear IHTerm_WT1. clear IHTerm_WT2.
rewrite Permuting_type_substitutions_cor1.
rewrite Permuting_type_substitutions_cor2.
replace (typeSubst_in_type T (S (S (length D2))) R) with T. trivial.
change (S (S (length D2))) with
       ((length ((classTypeParams C) :: Pi0 :: nil)) + (length D2)).
symmetry.
eapply Substitution_in_closed_type_aux.
destruct (method_declarations_are_well_formed m). apply H11.

(* T-new *)
intros.
change (typeSubst_in_term (Term_new N fts) (length D2) R) with
       (Term_new (typeSubst_in_type N (length D2) R) (typeSubst_in_term fts (length D2) R)).
change (typeSubst_in_termInfo (TermI N) (length D2) R) with
       (TermI (typeSubst_in_type N (length D2) R)).
replace (typeSubst_in_type N (length D2) R) with
       (Class_type C (typeSubst_in_type R0 (length D2) R)).
apply T_New.
rewrite <- typeSubst_of_class_type. rewrite H7 in H3. unfold N in H3.
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.

replace (classTypeParams C) with
        (typeSubst_in_type (classTypeParams C) (length D2) R).

(* rewrite length_of_append. rewrite length_of_typeSubst. *)
(* replace (lift_in_type (classTypeParams C) (length D1 + length D2)) with *)
(*         (typeSubst_in_type (lift_in_type (classTypeParams C) (length D)) (length D2) R). *)
(* *)
eapply Type_substitution_preserves_satisfaction. eauto. eauto. trivial.
rewrite <- H7. apply H4. rewrite <- H7.
(* rewrite <- H7. trivial. rewrite <- H7. *)
(* change (Type_WK (nil ++ D) (lift_in_type (classTypeParams C) (length D)) TypeSection). *)
replace D with (D ++ nil).
apply Weakening_of_well_kindedness_by_the_left.
destruct (classes_are_well_kinded C). trivial.
rewrite app_nil_end. trivial.
(* *)
change (length D2) with (length (nil (A := TypeS)) + (length D2)).
eapply Substitution_in_closed_type_aux.
destruct (classes_are_well_kinded C). apply H8.
(* *)
(* rewrite H7. rewrite length_of_append. simpl. *)
(* replace (length D1 + S (length D2)) with (S (length D2 + length D1)). *)
(* replace (length D1 + length D2) with (length D2 + length D1). *)
(* apply Type_lifting_stronger_than_substitution. omega. omega. *)
simpl in IHTerm_WT. unfold Class_type. apply IHTerm_WT. trivial.
rewrite Type_substitution_preserves_fields_of_fieldTermList. trivial.
trivial.

(* T-term-nil *)
intros. simpl. apply T_term_Nil.

(* T-term-cons *)
intros. simpl. apply T_term_Cons.
apply IHTerm_WT1; trivial.
apply IHTerm_WT2; trivial.

(* T-fieldTerm-nil *)
intros. simpl. apply T_fieldTerm_Nil.

(* T-fieldTerm-cons *)
rename S into T_t.
intros. simpl.
eapply T_fieldTerm_Cons with (R := (typeSubst_in_type R0 (length D2) R)).
(* *)
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
(* *)
fold C.
rewrite <- typeSubst_of_class_type.
eapply Type_substitution_preserves_subtyping. eauto. eauto. eauto. rewrite <- H7. trivial.
rewrite <- H7. trivial. rewrite <- H7. trivial.
(* *)
fold C. rewrite <- typeSubst_of_class_type.
eapply Type_substitution_preserves_well_kindedness. eauto.
eapply Satisfaction_implies_same_kind; eauto. eauto. trivial.
(* *)
simpl in IHTerm_WT1. apply IHTerm_WT1. trivial.
(* *)
fold T.
replace T with (typeSubst_in_type T (S (0 + length D2)) R).
rewrite <- Permuting_type_substitutions. simpl.
eapply Type_substitution_preserves_subtyping. eauto. eauto. eauto. rewrite <- H7. trivial.
eapply Typing_returns_well_kinded_types_cor1. rewrite <- H7. apply H3_.
trivial.
rewrite H7 in H5.
unfold T. eapply Substitution_in_field_type. eauto.
simpl.
change (S (length D2)) with ((length ((classTypeParams C) :: nil)) + (length D2)).
eapply Substitution_in_closed_type_aux.
destruct (field_declarations_are_well_formed f).
apply H8.
(* *)
apply IHTerm_WT2; trivial.
Qed.

(*****************************************************************)

Definition typeSubst_in_termEnv (G: TermEnv) (k: nat) (R: TypeS): TermEnv :=
  (map (fun T => (typeSubst_in_type T k R)) G)
.

Theorem Type_substitution_preserves_typing_cor1:
  forall (D: TypeEnv) (G: TermEnv) (Pi R T: TypeS) (k: KindS) (t: Term),
    (Type_WK D R (TypeTuple k)) ->
    (SectionSat D R Pi) ->
    let D' := (snoc D Pi) in
    (TypeEnv_WK D') ->
    (TermEnv_WK D' G) ->
    (Term_WT D' G t (TermI T)) ->
    (Term_WT D
             (typeSubst_in_termEnv G 0 R)
             (typeSubst_in_term t 0 R)
             (TermI (typeSubst_in_type T 0 R)))
.

Proof.
intros.
cut (
    (Term_WT (D ++ (typeSubst_in_typeEnv nil 0 R))
             (map (fun T => (typeSubst_in_type T (length (nil (A := TypeS))) R)) G)
             (typeSubst_in_term t (length (nil (A := TypeS))) R)
             (typeSubst_in_termInfo (TermI T) (length (nil (A := TypeS))) R))
).
simpl. rewrite <- app_nil_end. intros. trivial.
eapply Type_substitution_preserves_typing; eauto.
Qed.

Theorem Type_substitution_preserves_typing_cor2:
  forall (D: TypeEnv) (G: TermEnv) (Pi R T: TypeS) (k: KindS) (t: Term),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi) ->
    let D' := (Pi :: D) in
    (TypeEnv_WK D') ->
    (TermEnv_WK D' G) ->
    (Term_WT D' G t (TermI T)) ->
    (Term_WT (typeSubst_in_typeEnv D 0 R)
             (typeSubst_in_termEnv G (length D) R)
             (typeSubst_in_term t (length D) R)
             (TermI (typeSubst_in_type T (length D) R)))
.

Proof.
intros.
cut (
    (Term_WT (nil ++ (typeSubst_in_typeEnv D 0 R))
             (map (fun T => (typeSubst_in_type T (length D) R)) G)
             (typeSubst_in_term t (length D) R)
             (typeSubst_in_termInfo (TermI T) (length D) R))
).
simpl. intros. trivial.
eapply Type_substitution_preserves_typing; eauto.
Qed.

Theorem Type_substitution_preserves_typing_cor3:
  forall (G: TermEnv) (Pi Pi' R T: TypeS) (k: KindS) (t: Term),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi) ->
    let D := (Pi :: Pi' :: nil) in
    (TypeEnv_WK D) ->
    (TermEnv_WK D G) ->
    (Term_WT D G t (TermI T)) ->
    (Term_WT ((typeSubst_in_type Pi' 0 R) :: nil)
             (typeSubst_in_termEnv G 1 R)
             (typeSubst_in_term t 1 R)
             (TermI (typeSubst_in_type T 1 R)))
.

Proof.
intros.
cut (
    (Term_WT (typeSubst_in_typeEnv (Pi' :: nil) 0 R)
             (typeSubst_in_termEnv G (length (Pi' :: nil)) R)
             (typeSubst_in_term t (length (Pi' :: nil)) R)
             (TermI (typeSubst_in_type T (length (Pi' :: nil)) R)))
).
simpl. intros. trivial.
eapply Type_substitution_preserves_typing_cor2; eauto.
Qed.

Theorem Type_substitution_preserves_typing_cor4:
  forall (Us: TermEnv) (Pi Pi' R T U: TypeS) (k: KindS) (t: Term),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi) ->
    let D := (Pi :: Pi' :: nil) in
    let G := (U :: Us) in
    (TypeEnv_WK D) ->
    (TermEnv_WK D G) ->
    (Term_WT D G t (TermI T)) ->
    (Term_WT ((typeSubst_in_type Pi' 0 R) :: nil)
             ((typeSubst_in_type U 1 R) :: (typeSubst_in_termEnv Us 1 R))
             (typeSubst_in_term t 1 R)
             (TermI (typeSubst_in_type T 1 R)))
.

Proof.
intros.
cut (
    (Term_WT ((typeSubst_in_type Pi' 0 R) :: nil)
             (typeSubst_in_termEnv (U :: Us) 1 R)
             (typeSubst_in_term t 1 R)
             (TermI (typeSubst_in_type T 1 R)))

).
simpl. trivial.
eapply Type_substitution_preserves_typing_cor3; eauto.
Qed.

Theorem Type_substitution_preserves_typing_cor5:
  forall (G: TermEnv) (Pi R T: TypeS) (k: KindS) (t: Term),
    (Type_WK nil R (TypeTuple k)) ->
    (SectionSat nil R Pi) ->
    let D := (Pi :: nil) in
    (TypeEnv_WK D) ->
    (TermEnv_WK D G) ->
    (Term_WT D G t (TermI T)) ->
    (Term_WT nil
             (typeSubst_in_termEnv G 0 R)
             (typeSubst_in_term t 0 R)
             (TermI (typeSubst_in_type T 0 R)))
.

Proof.
intros.
cut (
    (Term_WT (nil ++ (typeSubst_in_typeEnv nil 0 R))
             (map (fun T => (typeSubst_in_type T (length (nil (A := TypeS))) R)) G)
             (typeSubst_in_term t (length (nil (A := TypeS))) R)
             (typeSubst_in_termInfo (TermI T) (length (nil (A := TypeS))) R))
).
simpl. intros. trivial.
eapply Type_substitution_preserves_typing; eauto.
Qed.


End SetProgram.
